import logging
import multiprocessing as mp
import os
import pickle
import subprocess
import timeit
from pathlib import Path

import pandas as pd
import parallel as par

import atom3.database as db
import atom3.sequence as sequ
from atom3.utils import extract_hmm_profile, slice_list


def add_conservation_parser(subparsers, pp):
    """Add parser."""

    def map_all_pssms_main(args):
        map_all_pssms(args.pkl_dataset, args.pruned_dataset, args.blastdb,
                      args.output_dir, args.c, args.source_type, args.rank, args.size)

    ap = subparsers.add_parser(
        'conservation', description='sequence conservation',
        help='compute sequence conservation features',
        parents=[pp])
    ap.set_defaults(func=map_all_pssms_main)
    ap.add_argument('pkl_dataset', metavar='pkl', type=str,
                    help='parsed dataset')
    ap.add_argument('blastdb', metavar='bdb', type=str,
                    help='blast database to do lookups on')
    ap.add_argument('output_dir', metavar='output', type=str,
                    help='directory to output to')
    ap.add_argument('-c', metavar='cpus', default=mp.cpu_count(), type=int,
                    help='number of cpus to use for processing (default:'
                         ' number processors available on current machine)')
    ap.add_argument('--source_type', metavar='complex_type', default='rcsb', type=str,
                    help='whether the source PDBs are for bound or unbound complexes (i.e. RCSB (e.g. DIPS) or DB5 complexes)')


def gen_protrusion_index(psaia_config_file, file_list_file):
    """Generate protrusion index for file list of PDB structures."""
    logging.info("PSAIA'ing {:}".format(file_list_file))
    _psaia(psaia_config_file, file_list_file)


def gen_pssm(pdb_filename, blastdb, output_filename):
    """Generate PSSM and PSFM from sequence."""
    pdb_name = db.get_pdb_name(pdb_filename)
    out_dir = os.path.dirname(output_filename)
    work_dir = os.path.join(out_dir, 'work')
    if not os.path.exists(work_dir):
        os.makedirs(work_dir, exist_ok=True)
    fasta_format = work_dir + "/{:}.fa"
    id_format = work_dir + "/{:}.cpkl"
    chains, chain_fasta_filenames, id_filenames = sequ.pdb_to_fasta(
        pdb_filename, fasta_format, id_format, True)

    pssms = []
    for chain, chain_fasta_filename, id_filename in \
            zip(chains, chain_fasta_filenames, id_filenames):
        basename = os.path.splitext(chain_fasta_filename)[0]
        pssm_filename = "{}.pssm".format(basename)
        blast_filename = "{}.blast".format(basename)
        clustal_filename = "{}.clustal".format(basename)
        al2co_filename = "{}.al2co".format(basename)
        if not os.path.exists(pssm_filename):
            logging.info("Blasting {:}".format(chain_fasta_filename))
            _blast(chain_fasta_filename, pssm_filename, blast_filename,
                   blastdb)

        if not os.path.exists(pssm_filename):
            logging.warning("No hits for {:}".format(chain_fasta_filename))
            # Create empty file.
            open(pssm_filename, 'w').close()

        if not os.path.exists(clustal_filename):
            logging.info("Converting {:}".format(blast_filename))
            _to_clustal(blast_filename, clustal_filename)

        if not os.path.exists(al2co_filename):
            logging.info("Al2co {:}".format(al2co_filename))
            _al2co(clustal_filename, al2co_filename)

        if os.stat(pssm_filename).st_size != 0:
            pssm = pd.read_csv(
                pssm_filename, skiprows=2, skipfooter=6, delim_whitespace=True,
                engine='python', usecols=range(20), index_col=[0, 1])
            pssm = pssm.reset_index()
            del pssm['level_0']
            pssm.rename(columns={'level_1': 'orig'}, inplace=True)

            pscm = pd.read_csv(
                pssm_filename, skiprows=2, skipfooter=6, delim_whitespace=True,
                engine='python', usecols=range(20, 40), index_col=[0, 1])
            psfm = pscm.applymap(lambda x: x / 100.)
            psfm = psfm.reset_index()
            del psfm['level_0']
            psfm.columns = pssm.columns
            del psfm['orig']
            del pssm['orig']

            # Combine both into one.
            psfm = psfm.add_prefix('psfm_')
            pssm = pssm.add_prefix('pssm_')
            al2co = pd.read_csv(
                al2co_filename, delim_whitespace=True, usecols=[2],
                names=['al2co'])
            pssm = pd.concat([pssm, psfm, al2co], axis=1)

        else:
            logging.warning("No pssm found for {:} (model {:}, chain {:})"
                            .format(pdb_name, chain[-2], chain[-1]))
            pssm, psfm = None, None

        pdb_name = db.get_pdb_name(pdb_filename)
        key = pdb_name + '-' + chain[-2] + '-' + chain[-1]
        pos_to_res = pickle.load(open(id_filename, 'rb'))[key]

        if pssm is not None:  # Skip if PSSM was not found
            pssm['pdb_name'] = db.get_pdb_name(pdb_filename)
            pssm['model'] = chain[0]
            pssm['chain'] = chain[1]
            pssm['residue'] = pos_to_res
            pssms.append(pssm)
    pssms = pd.concat(pssms)
    return pssms


def gen_profile_hmm(num_cpus, pkl_filename, output_filename, hhsuite_db, source_type, num_iter):
    """Generate profile HMM from sequence."""
    pdb_name = db.get_pdb_name(pkl_filename)
    out_dir = os.path.dirname(output_filename)
    work_dir = os.path.join(out_dir, 'work')
    if not os.path.exists(work_dir):
        os.makedirs(work_dir, exist_ok=True)
    fasta_format = work_dir + "/{:}.fa"
    id_format = work_dir + "/{:}.cpkl"

    # Get FASTA sequence-chain representations of PDB structures
    chains, chain_fasta_filenames, id_filenames = sequ.pdb_to_fasta(pkl_filename, fasta_format, id_format, True)

    # Process each profile HMM for a given PDB structure or complex
    num_chains = 0
    profile_hmms = []
    for chain, chain_fasta_filename, id_filename in zip(chains, chain_fasta_filenames, id_filenames):
        basename = os.path.splitext(chain_fasta_filename)[0]
        profile_hmm_filename = "{}.hhm".format(basename)
        hhblits_filename = "{}.a3m".format(basename)

        if not os.path.exists(profile_hmm_filename):
            logging.info("HHblits'ing {:}".format(chain_fasta_filename))
            _hhsuite(num_cpus, chain_fasta_filename, hhblits_filename, profile_hmm_filename, hhsuite_db, num_iter)

        if not os.path.exists(profile_hmm_filename):
            logging.warning("No hits for {:}".format(chain_fasta_filename))
            # Create empty file
            open(profile_hmm_filename, 'w').close()

        if os.stat(profile_hmm_filename).st_size != 0:
            with open(chain_fasta_filename, 'r') as fasta:
                with open(profile_hmm_filename, 'r') as hmm:
                    sequence = ''
                    for seq_line in fasta.readlines()[1:]:
                        sequence += " ".join(seq_line.splitlines())
                    profile_hmm = extract_hmm_profile(hmm.read(), sequence)
        else:
            logging.warning("No profile HMM found for {:} (model {:}, chain {:})"
                            .format(pdb_name, chain[-2], chain[-1]))
            profile_hmm = None

        pdb_name = db.get_pdb_name(pkl_filename)
        key = pdb_name + '-' + chain[-2] + '-' + chain[-1]
        pos_to_res = pickle.load(open(id_filename, 'rb'))[key]

        if profile_hmm is not None:  # Skip if profile HMM was not found
            profile_hmm = pd.DataFrame(data=profile_hmm)
            profile_hmm.insert(0, 'pdb_name', db.get_pdb_name(pkl_filename))
            profile_hmm.insert(1, 'model', chain[0])
            profile_hmm.insert(2, 'chain', chain[1])
            profile_hmm.insert(3, 'residue', pos_to_res)
            profile_hmms.append(profile_hmm)
        # Keep track of how many chains have been processed
        num_chains += 1
    # Merge related DataFrames into a single one
    profile_hmms = pd.concat(profile_hmms)
    return profile_hmms, num_chains


def map_protrusion_indices(psaia_config_file, file_list_file):
    start_time = timeit.default_timer()
    start_time_psaiaing = timeit.default_timer()
    gen_protrusion_index(psaia_config_file, file_list_file)
    elapsed_psaiaing = timeit.default_timer() - start_time_psaiaing

    start_time_writing = timeit.default_timer()
    elapsed_writing = timeit.default_timer() - start_time_writing

    elapsed = timeit.default_timer() - start_time
    logging.info(('For generating protrusion indices, spent {:05.2f} PSAIA\'ing,'
                  ' {:05.2f} writing, and {:05.2f} overall.').format(elapsed_psaiaing, elapsed_writing, elapsed))


def map_pssms(pdb_filename, blastdb, output_filename):
    pdb_name = db.get_pdb_name(pdb_filename)
    start_time = timeit.default_timer()
    start_time_blasting = timeit.default_timer()
    pis = gen_pssm(pdb_filename, blastdb, output_filename)
    num_chains = len(pis.groupby(['pdb_name', 'model', 'chain']))
    elapsed_blasting = timeit.default_timer() - start_time_blasting

    parsed = pd.read_pickle(pdb_filename)
    parsed = parsed.merge(pis, on=['model', 'pdb_name', 'chain', 'residue'])

    start_time_writing = timeit.default_timer()
    parsed.to_pickle(output_filename)
    elapsed_writing = timeit.default_timer() - start_time_writing

    elapsed = timeit.default_timer() - start_time
    logging.info(('For {:d} pssms generated from {}, spent {:05.2f} blasting,'
                  ' {:05.2f} writing, and {:05.2f} overall.').format(num_chains, pdb_name, elapsed_blasting,
                                                                     elapsed_writing, elapsed))


def map_profile_hmms(num_cpus, pkl_filename, output_filename, hhsuite_db, source_type, num_iter):
    pdb_name = db.get_pdb_name(pkl_filename)
    start_time = timeit.default_timer()
    start_time_blitsing = timeit.default_timer()
    profile_hmms, num_chains = gen_profile_hmm(num_cpus, pkl_filename, output_filename,
                                               hhsuite_db, source_type, num_iter)
    elapsed_blitsing = timeit.default_timer() - start_time_blitsing

    start_time_writing = timeit.default_timer()
    profile_hmms.to_pickle(output_filename)
    elapsed_writing = timeit.default_timer() - start_time_writing

    elapsed = timeit.default_timer() - start_time
    logging.info(('For {:d} profile HMMs generated from {}, spent {:05.2f} blitsing,'
                  ' {:05.2f} writing, and {:05.2f} overall.').format(num_chains, pdb_name, elapsed_blitsing,
                                                                     elapsed_writing, elapsed))


def _psaia(psaia_config_file, file_list_file):
    """Run PSAIA on specified input."""
    psaia_command = "yes y | psa {:} {:}"  # PSA is the CLI for PSAIA (the GUI)
    log_out = "{}.out".format(file_list_file)
    log_err = "{}.err".format(file_list_file)
    with open(log_out, 'a') as f_out:
        with open(log_err, 'a') as f_err:
            command = psaia_command.format(psaia_config_file, file_list_file)
            f_out.write('=================== CALL ===================\n')
            f_out.write(command + '\n')
            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
            f_out.write('================= END CALL =================\n')


def _hhsuite(num_cpus, query, output_a3m, output_hhm, hhsuite_db, num_iter):
    """Run HH-suite3 on specified input."""
    hhsuite_command = "hhblits -cpu {:} -i {:} -d {:} -oa3m {:} -n {} && hhmake -i {:} -o {:}"
    log_out = "{}.out".format(output_hhm)
    log_err = "{}.err".format(output_hhm)
    with open(log_out, 'a') as f_out:
        with open(log_err, 'a') as f_err:
            command = hhsuite_command.format(num_cpus, query, hhsuite_db, output_a3m, num_iter, output_a3m, output_hhm)
            f_out.write('=================== CALL ===================\n')
            f_out.write(command + '\n')
            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
            f_out.write('================= END CALL =================\n')


def _blast(query, output_pssm, output, blastdb):
    """Run PSIBlast on specified input."""
    psiblast_command = "psiblast -db {:} -query {:} -out_ascii_pssm {:} " + \
                       "-save_pssm_after_last_round -out {:}"
    log_out = "{}.out".format(output)
    log_err = "{}.err".format(output)
    with open(log_out, 'a') as f_out:
        with open(log_err, 'a') as f_err:
            command = psiblast_command.format(blastdb, query, output_pssm, output)
            f_out.write('=================== CALL ===================\n')
            f_out.write(command + '\n')
            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
            f_out.write('================= END CALL =================\n')


def _to_clustal(psiblast_in, clustal_out):
    """Convert PSIBlast output to CLUSTAL format."""
    log_out = "{}.out".format(clustal_out)
    log_err = "{}.err".format(clustal_out)
    mview_command = "mview -in blast -out clustal {:} | tail -n+4 > {:}"
    with open(log_out, 'a') as f_out:
        with open(log_err, 'a') as f_err:
            command = mview_command.format(psiblast_in, clustal_out)
            f_out.write('=================== CALL ===================\n')
            f_out.write(command + '\n')
            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
            f_out.write('================= END CALL =================\n')


def _al2co(clustal_in, al2co_out):
    """Use al2co on CLUSTAL format."""
    log_out = "{}.out".format(al2co_out)
    log_err = "{}.err".format(al2co_out)
    al2co_command = "al2co -i {:} -g 0.9 | head -n -12 > {:}"
    with open(log_out, 'a') as f_out:
        with open(log_err, 'a') as f_err:
            command = al2co_command.format(clustal_in, al2co_out)
            f_out.write('=================== CALL ===================\n')
            f_out.write(command + '\n')
            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
            f_out.write('================= END CALL =================\n')


def map_all_protrusion_indices(psaia_config_file, pdb_dataset, pkl_dataset, pruned_dataset, output_dir, source_type):
    ext = '.pkl'
    if source_type.lower() == 'rcsb':  # Filter out pairs that did not survive pruning previously to reduce complexity
        pruned_pdb_names = [db.get_pdb_name(filename)
                            for filename in db.get_structures_filenames(pruned_dataset, extension='.dill')]
        requested_filenames = [
            os.path.join(pkl_dataset, db.get_pdb_code(pruned_pdb_name)[1:3], pruned_pdb_name.split('_')[0] + ext)
            for pruned_pdb_name in pruned_pdb_names
        ]
    else:  # DB5 does not employ pair pruning, so there are no pairs to filter
        requested_filenames = [filename for filename in db.get_structures_filenames(pkl_dataset, extension=ext)]

    # Filter DB5 filenames to unbound type and get all work filenames
    requested_filenames = [filename for filename in requested_filenames
                           if (source_type.lower() == 'db5' and '_u_' in filename)
                           or (source_type.lower() == 'rcsb')]
    requested_keys = [db.get_pdb_name(x) for x in requested_filenames]
    requested_pdb_codes = [db.get_pdb_code(x) for x in requested_filenames]
    produced_filenames_path = os.path.join(output_dir, 'PSAIA', 'RCSB' if source_type.lower() == 'rcsb' else 'DB5')
    produced_filenames = [path.as_posix() for path in Path(produced_filenames_path).rglob('*.tbl')]
    produced_keys = [db.get_pdb_code(x) for x in produced_filenames]
    work_keys = [key for key, pdb_code in zip(requested_keys, requested_pdb_codes) if pdb_code not in produced_keys]
    if source_type.lower() == 'rcsb':
        work_filenames = [os.path.join(pdb_dataset, db.get_pdb_code(work_key)[1:3], work_key)
                          for work_key in work_keys]
    else:
        work_filenames = [os.path.join(pdb_dataset, db.get_pdb_code(work_key).upper(), work_key)
                          for work_key in work_keys]

    # Exit early if no inputs need to processed
    num_inputs = len(work_filenames)
    if num_inputs == 0:
        logging.info("Exiting early since all provided PDB files have already been processed by PSAIA")
        exit(0)

    # Create comprehensive filename list for PSAIA to single-threadedly process for requested features (e.g. protrusion)
    file_list_file = os.path.join(output_dir, 'PSAIA', source_type.upper(), 'pdb_list.fls')
    with open(file_list_file, 'w') as file:
        for requested_pdb_filename in work_filenames:
            file.write(f'{requested_pdb_filename}\n')

    logging.info("{:} PDB files to process with PSAIA".format(num_inputs))
    inputs = [(psaia_config_file, file_list_file)]
    par.submit_jobs(map_protrusion_indices, inputs, 1)  # PSAIA is inherently single-threaded in execution


def map_all_pssms(pkl_dataset, pruned_dataset, blastdb, output_dir, num_cpus, source_type, rank, size):
    ext = '.pkl'
    if source_type.lower() == 'rcsb':  # Filter out pairs that did not survive pruning previously to reduce complexity
        pruned_pdb_names = [db.get_pdb_name(filename)
                            for filename in db.get_structures_filenames(pruned_dataset, extension='.dill')]
        requested_filenames = [
            os.path.join(pkl_dataset, db.get_pdb_code(pruned_pdb_name)[1:3], pruned_pdb_name.split('_')[0] + ext)
            for pruned_pdb_name in pruned_pdb_names
        ]
    else:  # DB5 does not employ pair pruning, so there are no pairs to filter
        requested_filenames = [filename for filename in db.get_structures_filenames(pkl_dataset, extension=ext)]

    # Filter DB5 filenames to unbound type and get all work filenames
    requested_filenames = [filename for filename in requested_filenames
                           if (source_type.lower() == 'db5' and '_u_' in filename)
                           or (source_type.lower() == 'rcsb')]
    requested_keys = [db.get_pdb_name(x) for x in requested_filenames]
    produced_filenames = db.get_structures_filenames(output_dir, extension='.pkl')
    produced_keys = [db.get_pdb_name(x) for x in produced_filenames]
    work_keys = [key for key in requested_keys if key not in produced_keys]
    if source_type.lower() == 'rcsb':
        work_filenames = [os.path.join(pkl_dataset, db.get_pdb_code(work_key)[1:3], work_key + ext)
                          for work_key in work_keys]
    else:
        work_filenames = [os.path.join(pkl_dataset, db.get_pdb_code(work_key)[1:3].upper(), work_key + ext)
                          for work_key in work_keys]

    # Reserve an equally-sized portion of the full work load for a given rank in the MPI world
    work_filename_rank_batches = slice_list(work_filenames, size)
    work_filenames = work_filename_rank_batches[rank]

    output_filenames = []
    for pdb_filename in work_filenames:
        sub_dir = output_dir + '/' + db.get_pdb_code(pdb_filename)[1:3]
        if not os.path.exists(sub_dir):
            os.makedirs(sub_dir, exist_ok=True)
        output_filenames.append(sub_dir + '/' + db.get_pdb_name(pdb_filename) + ".pkl")

    logging.info("{:} requested keys, {:} produced keys, {:} work keys"
                 .format(len(requested_keys), len(produced_keys),
                         len(work_keys)))
    inputs = [(key, blastdb, output)
              for key, output in zip(work_filenames, output_filenames)]
    par.submit_jobs(map_pssms, inputs, num_cpus)


def map_all_profile_hmms(pkl_dataset, pruned_dataset, output_dir, hhsuite_db, num_cpu_jobs,
                         num_cpus_per_job, source_type, num_iter, rank, size, write_file):
    ext = '.pkl'
    if write_file:
        if source_type.lower() == 'rcsb':  # Filter out pairs that did not survive pruning previously to reduce complexity
            pruned_pdb_names = [db.get_pdb_name(filename)
                                for filename in db.get_structures_filenames(pruned_dataset, extension='.dill')]
            requested_filenames = [
                os.path.join(pkl_dataset, db.get_pdb_code(pruned_pdb_name)[1:3], pruned_pdb_name.split('_')[0] + ext)
                for pruned_pdb_name in pruned_pdb_names
            ]
        else:  # DB5 does not employ pair pruning, so there are no pairs to filter
            requested_filenames = [filename for filename in db.get_structures_filenames(pkl_dataset, extension=ext)]

        # Filter DB5 filenames to unbound type and get all work filenames
        requested_filenames = [filename for filename in requested_filenames
                               if (source_type.lower() == 'db5' and '_u_' in filename)
                               or (source_type.lower() == 'rcsb')]
        requested_keys = [db.get_pdb_name(x) for x in requested_filenames]
        produced_filenames = db.get_structures_filenames(output_dir, extension='.pkl')
        produced_keys = [db.get_pdb_name(x) for x in produced_filenames]
        work_keys = [key for key in requested_keys if key not in produced_keys]
        if source_type.lower() == 'rcsb':
            work_filenames = [os.path.join(pkl_dataset, db.get_pdb_code(work_key)[1:3], work_key + ext)
                              for work_key in work_keys]
        else:
            work_filenames = [os.path.join(pkl_dataset, db.get_pdb_code(work_key)[1:3].upper(), work_key + ext)
                              for work_key in work_keys]

        logging.info("{:} requested keys, {:} produced keys, {:} work keys".format(len(requested_keys),
                                                                                   len(produced_keys), len(work_keys)))

        # Write out a local file containing all work filenames
        work_filenames = list(set(work_filenames))  # Remove any duplicate filenames
        temp_df = pd.DataFrame({'filename': work_filenames})
        temp_df.to_csv(f'{source_type}_work_filenames.csv')
        logging.info('File containing work filenames written to storage. Exiting...')

    # Read from previously-created work filenames CSV
    else:
        work_filenames = pd.read_csv(f'{source_type}_work_filenames.csv').iloc[:, 1].to_list()
        work_filenames = list(set(work_filenames))  # Remove any duplicate filenames

        # Reserve an equally-sized portion of the full work load for a given rank in the MPI world
        work_filename_rank_batches = slice_list(work_filenames, size)
        work_filenames = work_filename_rank_batches[rank]

        output_filenames = []
        for pdb_filename in work_filenames:
            sub_dir = output_dir + '/' + db.get_pdb_code(pdb_filename)[1:3]
            if not os.path.exists(sub_dir):
                os.makedirs(sub_dir, exist_ok=True)
            output_filenames.append(sub_dir + '/' + db.get_pdb_name(pdb_filename) + '.pkl')

        logging.info("{:} work filenames".format(len(work_filenames)))

        inputs = [(num_cpus_per_job, key, output, hhsuite_db, source_type, num_iter)
                  for key, output in zip(work_filenames, output_filenames)]
        par.submit_jobs(map_profile_hmms, inputs, num_cpu_jobs)



import collections as col
import logging
import os

import dill

import atom3.case as ca
import atom3.database as db

Complex = col.namedtuple(
    'Complex', ['name', 'bound_filenames', 'unbound_filenames'])


def add_complexes_parser(subparsers, pp):
    """Add parser."""

    def complexes_main(args):
        complexes(args.pkl_dataset, args.output_dill, args.type)

    ap = subparsers.add_parser(
        'complex', description='pickled file mapping to complexes',
        help='maps protein complexes to their individual subunits',
        parents=[pp])
    ap.set_defaults(func=complexes_main)
    ap.add_argument('pkl_dataset', type=str, help='pkl dataset')
    ap.add_argument('output_dill', metavar='complexes.dill', type=str,
                    help='dill file to output complexes to')
    ap.add_argument(
        '-t', '--type', dest='type',
        choices=['rcsb', 'db5', 'dockground', 'db5mut', 'hotspot'],
        default='rcsb', help='type of dataset (default: rcsb)')


def complexes(pkl_dataset, output_dill, type):
    if os.path.exists(output_dill):
        logging.warning("Complex file {:} already exists!".format(output_dill))
        return
    logging.info("Getting filenames...")
    filenames = db.get_structures_filenames(pkl_dataset, extension='.pkl')
    logging.info("Getting complexes...")
    complexes = get_complexes(filenames, type)
    write_complexes(complexes, output_dill)


def get_complexes(filenames, type):
    """
    Return mapping to files from same complex.

    If unbound true, will return a pair of list of files.  First one is bound
    files, second is their unbound counterpart.
    """
    if type == 'rcsb':
        raw_complexes = _get_rcsb_complexes(filenames)
    elif type == 'db5':
        raw_complexes = _get_db5_complexes(filenames)
    elif type == 'db5mut':
        raw_complexes = _get_db5_complexes_mut(filenames)
    elif type == 'hotspot':
        raw_complexes = _get_db5_complexes_hotspot_mut(filenames)
    elif type == 'dockground':
        raw_complexes = _get_db5_complexes(
            filenames, keyer=db.get_pdb_basename)
    else:
        raise RuntimeError("Unrecognized dataset type {:}".format(type))
    complexes = {
        'data': raw_complexes,
        'type': type
    }
    return complexes


def read_complexes(input_dill):
    with open(input_dill, 'rb') as f:
        complexes = dill.load(f)
    return complexes


def write_complexes(complexes, output_dill):
    if os.path.exists(output_dill):
        raise RuntimeError(
            "Complex file {:} already exists!".format(output_dill))
    if not os.path.exists(os.path.dirname(output_dill)):
        os.makedirs(os.path.dirname(output_dill), exist_ok=True)
    with open(output_dill, 'wb') as f:
        dill.dump(complexes, f)


def _get_rcsb_complexes(filenames):
    """Get complexes for RCSB type dataset."""
    complexes = {}
    for filename in filenames:
        name = db.get_pdb_name(filename)
        complexes[name] = Complex(name=name, bound_filenames=[filename],
                                  unbound_filenames=[])
    return complexes


def _get_db5_complexes_mut(filenames):
    filenames = [f for f in filenames if 'mut' in f] #screen the original non mutant
    basenames = [os.path.basename(f) for f in filenames]
    new_pdbs = [b[:4] + '_' + b[b.rfind('_') + 1: b.find('.')] for b in basenames]
    complexes = {}
    for pdb in new_pdbs:
        code4 = pdb[:4]
        mutnum = 'mut_' + pdb[pdb.find('_') + 1: ] + '.'
        lb = [x for x in filenames if code4 in x and mutnum in x and '_l_b_' in x][0]
        rb = [x for x in filenames if code4 in x and mutnum in x and '_r_b_' in x][0]
        lu = [x for x in filenames if code4 in x and mutnum in x and '_l_u_' in x][0]
        ru = [x for x in filenames if code4 in x and mutnum in x and '_r_u_' in x][0]
        complexes[pdb] = Complex(name=pdb, bound_filenames=[lb, rb], unbound_filenames=[lu, ru])
    return complexes


def _get_db5_complexes_hotspot_mut(filenames):
    filenames = sorted(filenames)
    complexes = {}
    file_pairs = {}
    for f in filenames:
        pdb = ''
        if 'mut' in f:
            pdb = os.path.basename(f)[:4] + f[f.rfind('_') + 1: f.find('.')]
        else:
            pdb = os.path.basename(f)[:4]
        if pdb in file_pairs:
            file_pairs[pdb] += [f]
        else:
            file_pairs[pdb] = [f]
    for pdb in file_pairs:
        if '_l_' in file_pairs[pdb][0]:
            complexes[pdb] = Complex(name=pdb, bound_filenames=file_pairs[pdb], unbound_filenames=[None, None])
        else:
            complexes[pdb] = Complex(name=pdb, bound_filenames=[file_pairs[pdb][1], file_pairs[pdb][0]], unbound_filenames=[None, None])
    return complexes


def _get_db5_complexes(filenames, keyer=db.get_pdb_code):
    """Get complexes for docking benchmark 5 type dataset."""
    pdb_codes = ca.get_complex_pdb_codes(filenames)
    complexes = {}
    for pdb_code in pdb_codes:
        lb = ca.find_of_type(
            pdb_code, filenames, receptor=False, bound=True, style='db5')
        rb = ca.find_of_type(
            pdb_code, filenames, receptor=True, bound=True, style='db5')
        lu = ca.find_of_type(
            pdb_code, filenames, receptor=False, bound=False, style='db5')
        ru = ca.find_of_type(
            pdb_code, filenames, receptor=True, bound=False, style='db5')
        if lb is None or rb is None:
            logging.warning("Skipping {:} since not all bound files present."
                            .format(pdb_code))
            continue
        if lu is None or ru is None:
            logging.warning(
                "Skipping {:} since not all unbound files present."
                .format(pdb_code))
            continue
        if keyer(lu) != keyer(lb) or keyer(lu) != keyer(rb) or \
                keyer(lu) != keyer(ru):
            logging.warning(
                "Skipping {:} since not all keys match."
                .format(pdb_code))
            continue

        complexes[keyer(lu)] = Complex(
            name=keyer(lu), bound_filenames=[lb, rb],
            unbound_filenames=[lu, ru])
    return complexes



import logging
import multiprocessing as mp
import os

import parallel as par

import atom3.database as db
import atom3.structure as struct


def add_parse_parser(subparsers, pp):
    """Add parser."""

    def parse_all_main(args):
        parse_all(args.pdb_dataset, args.output_dir, args.c)

    ap = subparsers.add_parser(
        'parse', description='pdb to dataframe',
        help='pdb dataset to pickled dataframes',
        parents=[pp])
    ap.set_defaults(func=parse_all_main)
    ap.add_argument('pdb_dataset', metavar='pdb', type=str,
                    help='pdb dataset')
    ap.add_argument('output_dir', metavar='output', type=str,
                    help='directory to output to')
    ap.add_argument('-c', metavar='cpus', default=mp.cpu_count(), type=int,
                    help='number of cpus to use for processing (default:'
                    ' number processors available on current machine)')


def parse_all(pdb_dataset, output_dir, num_cpus):
    """Parse pdb dataset (pdb files) to pandas dataframes."""
    requested_filenames = db.get_structures_filenames(pdb_dataset)
    produced_filenames = db.get_structures_filenames(
        output_dir, extension='.pkl')

    requested_keys = [db.get_pdb_name(x) for x in requested_filenames]
    produced_keys = [db.get_pdb_name(x) for x in produced_filenames]
    work_keys = [key for key in requested_keys if key not in produced_keys]
    work_filenames = [x[0] for x in
                      db.get_all_filenames(work_keys, pdb_dataset, enforcement=2)]

    logging.info("{:} requested keys, {:} produced keys, {:} work keys"
                 .format(len(requested_keys), len(produced_keys),
                         len(work_keys)))

    output_filenames = []
    for pdb_filename in work_filenames:
        sub_dir = output_dir + '/' + db.get_pdb_code(pdb_filename)[1:3]
        if not os.path.exists(sub_dir):
            os.makedirs(sub_dir, exist_ok=True)
        output_filenames.append(
            sub_dir + '/' + db.get_pdb_name(pdb_filename) + ".pkl")

    inputs = [(key, output)
              for key, output in zip(work_filenames, output_filenames)]
    par.submit_jobs(parse, inputs, num_cpus)


def parse(pdb_filename, output_pkl):
    """Parse single pdb file to output directory."""
    # Get middle two letters of code.
    logging.info("Reading {:}".format(pdb_filename))
    df = struct.parse_structure(pdb_filename, one_model=False)
    logging.info("Writing {:} to {:}".format(pdb_filename, output_pkl))
    df.to_pickle(output_pkl)
    logging.info("Done writing {:} to {:}"
                 .format(pdb_filename, output_pkl))



"""A full test case consisting of a bound and matching unbound complex."""
import logging
import multiprocessing as mp
import os
import pickle
import tempfile

import Bio.Blast.Applications as app
import Bio.PDB
import Bio.SVDSuperimposer as svd
import Bio.SeqIO
import numpy as np
import parallel as par

import atom3.database as db
import atom3.sequence as sequ
import atom3.structure as struct


def add_clean_parser(subparsers, pp):
    """Add parser."""

    def generate_all_clean_complexes_main(args):
        generate_all_clean_complexes(args)

    cp = subparsers.add_parser('clean', description="DB5 cleaning",
                               help='generate clean DB5 files',
                               parents=[pp])
    cp.set_defaults(func=generate_all_clean_complexes_main)
    cp.add_argument('pdb_dataset', type=str,
                    help='pdb file or dir to clean.')
    cp.add_argument('output_dir', type=str,
                    help='where to write cleaned output files.')
    cp.add_argument(
        '-s', '--style', dest='style', choices=['db5', 'dockground'],
        default='db5', help='style of dataset (default: db5)')
    cp.add_argument('-c', metavar='cpus', default=mp.cpu_count(), type=int,
                    help='number of cpus to use for processing (default:'
                    ' number processors available on current machine)')


def generate_all_clean_complexes(args):
    """Clean all complexes in input_dir, writing them out to output_dir."""
    requested_keys = get_complex_pdb_codes(args.pdb_dataset)
    produced_filenames = db.get_structures_filenames(args.output_dir)
    produced_keys = []
    for pdb_code in requested_keys:
        res = get_files_for_complex(pdb_code, produced_filenames, 'db5')
        if len([x for x in res if x is None]) == 0:
            produced_keys.append(pdb_code)
    work_keys = [key for key in requested_keys if key not in produced_keys]

    logging.info("{:} requested keys, {:} produced keys, {:} work keys"
                 .format(len(requested_keys), len(produced_keys),
                         len(work_keys)))

    inputs = [(pc, args.pdb_dataset, args.output_dir + '/' + pc, args.style)
              for pc in work_keys]
    par.submit_jobs(_generate_clean_complex, inputs, args.c)


def get_complex_pdb_codes(pdb_dataset):
    """Get complexes in provided directory."""
    complexes = set()
    for structure in db.get_structures_filenames(pdb_dataset):
        complexes.add(db.get_pdb_code(structure))
    return complexes


def get_files_for_complex(pdb_code, pdb_dataset, style):
    """Get case files corresponding to provided pdb_code."""
    lu = find_of_type(
        pdb_code, pdb_dataset, receptor=False, bound=False, style=style)
    ru = find_of_type(
        pdb_code, pdb_dataset, receptor=True, bound=False, style=style)
    if style == 'db5':
        lb = find_of_type(
            pdb_code, pdb_dataset, receptor=False, bound=True, style=style)
        rb = find_of_type(
            pdb_code, pdb_dataset, receptor=True, bound=True, style=style)
    elif style == 'dockground':
        b = find_of_type(
            pdb_code, pdb_dataset, receptor=None, bound=True, style=style)
        lb = b
        rb = b
    return lu, lb, ru, rb


def find_of_type(pdb_name_query, pdb_dataset, receptor, bound, style):
    """Get matching partner of provided file."""
    pdb_code = db.get_pdb_code(pdb_name_query)
    results = None
    for pdb_name in db.get_structures_filenames(pdb_dataset):
        if db.get_pdb_code(pdb_name) == pdb_code:
            if is_of_type(pdb_name, style, receptor=receptor, bound=bound):
                results = pdb_name
    return results


def _has_symbol(symbol, name):
    """
    Check if has provided symbol in name.

    Recognizes either the _SYMBOL pattern at end of string, or _SYMBOL_ in
    middle.
    """
    return name.endswith('_' + symbol) or ('_' + symbol + '_') in name


def is_of_type(pdb_name, style, receptor=None, bound=None):
    """Check if pdb_name is of requested type."""
    pdb_name = db.get_pdb_name(pdb_name, with_type=False)
    if receptor is None:
        if bound is None:
            return True
        elif bound:
            if style == 'db5':
                return _has_symbol('b', pdb_name)
            elif style == 'dockground':
                # Dockground only has pdb code.
                return len(pdb_name) == 4
        else:
            return _has_symbol('u', pdb_name)

    if bound is None:
        if receptor:
            return _has_symbol('r', pdb_name) or _has_symbol('2', pdb_name)
        else:
            return _has_symbol('l', pdb_name) or _has_symbol('1', pdb_name)

    if receptor and bound:
        return _has_symbol('r_b', pdb_name) or _has_symbol('2_b', pdb_name)
    elif receptor and not bound:
        return _has_symbol('r_u', pdb_name) or _has_symbol('2_u', pdb_name)
    elif not receptor and bound:
        return _has_symbol('l_b', pdb_name) or _has_symbol('1_b', pdb_name)
    else:
        return _has_symbol('l_u', pdb_name) or _has_symbol('1_u', pdb_name)

    return False


def ligand_bound(pdb_code):
    """Augment pdb code with ligand partner and bound binding notation."""
    return pdb_code + '_l_b'


def ligand_unbound(pdb_code):
    """Augment pdb code with ligand partner and unbound binding notation."""
    return pdb_code + '_l_u'


def receptor_bound(pdb_code):
    """Augment pdb code with receptor partner and bound binding notation."""
    return pdb_code + '_r_b'


def receptor_unbound(pdb_code):
    """Augment pdb code with receptor partner and unbound binding notation."""
    return pdb_code + '_r_u'


def get_partner_regex(pdb_name, bound):
    """
    Get partner regex for provided filename.

    e.g. ligand 11as_r_u_cleaned.pdb1 would give 11as_._u_cleaned.pdb1.
    """
    new_pdb_name = list(pdb_name)
    new_pdb_name[5] = '.'
    new_pdb_name[7] = 'b' if bound else 'u'
    new_pdb_name = ['^'] + new_pdb_name + ['$']
    return "".join(new_pdb_name)


def get_binder_regex(pdb_name, receptor):
    """
    Get binder regex for provided filename.

    e.g. bound 11as_r_u_cleaned.pdb1 would give 11as_r_._cleaned.pdb1.
    """
    new_pdb_name = list(pdb_name)
    new_pdb_name[7] = '.'
    new_pdb_name[5] = 'r' if receptor else 'l'
    new_pdb_name = ['^'] + new_pdb_name + ['$']
    return "".join(new_pdb_name)


def _get_binding(pdb_filename):
    """
    Get binding mode of pdb file.  x if none.

    e.g. 11as_r_u.pdb would give u
    """
    if "_u" in pdb_filename:
        return "u"
    elif "_b" in pdb_filename:
        return "b"
    else:
        return "x"


def _get_partner(pdb_filename):
    """
    Get binding mode of pdb file.  x if none.

    e.g. 11as_r_u.pdb would give u
    """
    if "_r" in pdb_filename or "_2" in pdb_filename:
        return "r"
    elif "_l" in pdb_filename or "_1" in pdb_filename:
        return "l"
    else:
        return "x"


def get_pdb_code_with_partner(pdb_filename):
    """
    Get pdb code with partner annotated.

    e.g. 11as_r_u.pdb would give 11as_r
    """
    return db.get_pdb_code(pdb_filename) + '_' + _get_partner(pdb_filename)


def get_pdb_code_with_partner_and_binding(pdb_filename):
    """
    Get pdb code with partner and binding state annotated.

    e.g. 11as_r_u.pdb would give 11as_r_u
    """
    return db.get_pdb_code(pdb_filename) + '_' + _get_partner(pdb_filename) + \
        '_' + _get_binding(pdb_filename)


def get_pdb_code_with_binding(pdb_filename):
    """
    Get pdb code with binding state annotated.

    e.g. 11as_r_u.pdb would give 11as_u
    """
    return db.get_pdb_code(pdb_filename) + '_' + _get_binding(pdb_filename)


def _tmp_fasta(seq):
    """Output provided sequences as temporaray FASTA file."""
    output = tempfile.NamedTemporaryFile().name
    sequ.write_fasta({output: seq}, output)
    return output


def _get_seq_and_atoms(filename):
    """Form dictionaries mapping from chain to their sequence and atoms."""
    seqs = {}
    all_atoms = {}
    structure = struct.parse_structure(filename)
    pdb_name = db.get_pdb_name(filename)
    for (chain, residues) in \
            struct.get_chain_to_valid_residues(structure, pdb_name):
        atoms = []
        for residue in residues:
            atoms.append(np.array(residue[['x', 'y', 'z']], dtype='f4'))
        if len(residues) != 0:
            # Ignore zero-length peptides.
            seqs[chain] = residues
            all_atoms[chain] = np.array(atoms)
    return all_atoms, seqs


def _get_alignments(seq_u, atom_u, seq_b, atom_b):
    """
    Get alignment between two chains.

    Does structural and sequence alignment.
    """
    fasta_u = _tmp_fasta(seq_u)
    fasta_b = _tmp_fasta(seq_b)

    # Only use small word size for small sequences.
    # TODO. This is a bit brittle.
    word_size = 2 if len(seq_u) < 10 and len(seq_b) < 10 else 3

    blastp_cline = app.NcbiblastpCommandline(
        subject=fasta_u, query=fasta_b,
        outfmt="'10 qstart qend sstart send qseq sseq ppos evalue'",
        num_alignments=1, word_size=word_size, culling_limit=1, evalue=0.1)
    out, err = blastp_cline()

    # The trailing token is empty.
    alignments = [x for x in out.split('\n') if x]
    b2r, u2r = {}, {}
    b2u, u2b = {}, {}
    aligned_pos_b, aligned_pos_u = [], []
    all_ppos = []
    if len(out) == 0:
        # No hits found.
        return 0.0, float('inf'), (None, None)

    warned = False
    if len(alignments) > 1:
        logging.warning("More than one alignment found.")
    for i, curr in enumerate(alignments):
        start_b, end_b, start_u, end_u, align_b, align_u, ppos, evalue = \
            curr.split(',')
        start_b, end_b = int(start_b), int(end_b)
        start_u, end_u = int(start_u), int(end_u)
#        logging.info('Alignment {:} (score {:}) from {:} to {:} on bound, '
#                     '{:} to {:} on unbound.'.format(
#                         i, evalue, start_b, end_b, start_u, end_u))
        idx_b, idx_u = start_b - 1, start_u - 1
        assert len(align_u) == len(align_b)
        align_size = len(align_u)
        for k in range(align_size):
            if align_b[k] != '-' and align_u[k] != '-':
                if idx_b not in b2u and idx_u not in u2b:
                    b2u[idx_b] = idx_u
                    u2b[idx_u] = idx_b
                    aligned_pos_b.append(idx_b)
                    aligned_pos_u.append(idx_u)
                else:
                    if not warned:
                        logging.warning('ignoring double prediction {:} bound '
                                        'to {:} unbound'.format(idx_u, idx_b))
                        logging.warning('not showing future warnings for this '
                                        'alignment')
                        warned = True
            if align_u[k] != '-':
                idx_u += 1
            if align_b[k] != '-':
                idx_b += 1
        all_ppos.append((align_size, float(ppos)))

    idx_u, idx_b = 0, 0
    idx_r = 1
    u2r, b2r = {}, {}

    while idx_u != len(seq_u) or idx_b != len(seq_b):
        if idx_u in u2b and idx_b in b2u:
            u2r[idx_u] = idx_r
            b2r[idx_b] = idx_r
            idx_u += 1
            idx_b += 1
        elif idx_u not in u2b and idx_u != len(seq_u):
            u2r[idx_u] = idx_r
            idx_u += 1
        elif idx_b not in b2u and idx_b != len(seq_b):
            b2r[idx_b] = idx_r
            idx_b += 1
        idx_r += 1

    total = 0
    total_ppos = 0
    for align_size, ppos in all_ppos:
        total_ppos += ppos * align_size
        total += align_size
    total_ppos /= total

    sup = svd.SVDSuperimposer()
    sup.set(atom_u[aligned_pos_u], atom_b[aligned_pos_b])
    return total_ppos, sup.get_init_rms(), (b2r, u2r)


def _compute_similarities(unbound, bound):
    """Get all possible alignments between chains of two PDB files."""
    atoms_u, seqs_u = _get_seq_and_atoms(unbound)
    atoms_b, seqs_b = _get_seq_and_atoms(bound)
    alignments = {}
    chains_u, chains_b = seqs_u.keys(), seqs_b.keys()
    num_u, num_b = len(seqs_u), len(seqs_b)
    seq_sims = np.zeros((num_u, num_b))
    struct_sims = np.zeros((num_u, num_b))
    for i, chain_u in enumerate(chains_u):
        for j, chain_b in enumerate(chains_b):
            # logging.info("unbound chain {:} vs. bound chain {:}"
            #              .format(chain_u, chain_b))
            seq_sims[i, j], struct_sims[i, j], alignments[(i, j)] = \
                _get_alignments(seqs_u[chain_u], atoms_u[chain_u],
                                seqs_b[chain_b], atoms_b[chain_b])
    return chains_b, chains_u, seq_sims, struct_sims, alignments


def _get_chain_mapping(unbound, bound, style):
    """Based on alignments, get mapping from chain to chain and res to res."""
    chains_b, chains_u, seq_sims, struct_sims, alignments = \
        _compute_similarities(unbound, bound)

    b2u_chain = {}
    u2b_chain = {}
    b2r_res = {}
    u2r_res = {}
    for j, chain_b in enumerate(chains_b):
        best_chain = None
        best_seq_score = 0
        best_struct_score = float('inf')
        best_alignment = None
        for i, chain_u in enumerate(chains_u):
            curr_seq_score = seq_sims[i, j]
            curr_struct_score = struct_sims[i, j]
            if best_chain is None and curr_seq_score != 0.0:
                best_chain = chain_u
                best_seq_score = curr_seq_score
                best_struct_score = curr_struct_score
                best_alignment = alignments[(i, j)]
            elif curr_seq_score > best_seq_score - 20 and \
                    curr_struct_score < best_struct_score:
                # If the sequence similarity score is comparable and structural
                # distance is better.
                best_chain = chain_u
                best_seq_score = curr_seq_score
                best_struct_score = curr_struct_score
                best_alignment = alignments[(i, j)]
        if best_chain is not None:
            # For DB5, we assume that each bound chain has a matching unbound,
            # but not the reverse.
            b2u_chain[chain_b] = best_chain
            u2b_chain[best_chain] = chain_b
            b2r_res[chain_b] = best_alignment[0]
            u2r_res[best_chain] = best_alignment[1]

    best_seq_match_for_each_bound = np.max(seq_sims, axis=0)
    # If the sequence similarity score is comparable, we rely on structural
    # distance.
    seq_threshold_for_each_bound = best_seq_match_for_each_bound - 20
    candidates_for_each_bound = seq_sims > seq_threshold_for_each_bound

    best_seq_match_for_each_unbound = np.max(seq_sims, axis=1)
    # If the sequence similarity score is comparable, we rely on structural
    # distance.
    seq_threshold_for_each_unbound = best_seq_match_for_each_unbound - 20
    candidates_for_each_unbound = seq_sims.T > seq_threshold_for_each_unbound

    b2u_proposed = {}
    u2b_proposed = {}
    for i, bound_candidates in enumerate(candidates_for_each_unbound.T):
        if np.max(seq_sims[i, :]) == 0.0:
            continue
        # All bound sorted by how structurally similar they are to unbound.
        struct_idx = np.argsort(struct_sims[i, :])
        # Choose best bound by finding best structure that has OK sequence
        # match.
        best_bound = struct_idx[np.argmax(bound_candidates[struct_idx])]
        u2b_proposed[i] = best_bound
    for i, unbound_candidates in enumerate(candidates_for_each_bound.T):
        if np.max(seq_sims[:, i]) == 0.0:
            continue
        # All unbound sorted by how structurally similar they are to bound.
        struct_idx = np.argsort(struct_sims[:, i])
        # Choose best bound by finding best structure that has OK sequence
        # match.
        best_unbound = struct_idx[np.argmax(unbound_candidates[struct_idx])]
        b2u_proposed[i] = best_unbound

    b2u_chain_proposed = {}
    for b, u in b2u_proposed.items():
        b2u_chain_proposed[chains_b[b]] = chains_u[u]
    u2b_chain_proposed = {}
    for u, b in u2b_proposed.items():
        u2b_chain_proposed[chains_u[u]] = chains_b[b]

    if style == 'db5':
        # For DB5, we assume that each bound chain has a matching unbound, but
        # not the reverse.
        assert u2b_chain_proposed == u2b_chain
        assert b2u_chain_proposed == b2u_chain
        u2b_chain, b2u_chain = {}, {}
        for b, u in b2u_proposed.items():
            chain_b, chain_u = chains_b[b], chains_u[u]
            b2u_chain[chain_b] = chain_u
            u2b_chain[chain_u] = chain_b
            b2r_res[chain_b] = alignments[u, b][0]
            u2r_res[chain_u] = alignments[u, b][1]
    elif style == 'dockground':
        # For dockground, we assume that each unbound chain has a matching
        # bound, but not the reverse.
        u2b_chain, b2u_chain = {}, {}
        for u, b in u2b_proposed.items():
            chain_b, chain_u = chains_b[b], chains_u[u]
            b2u_chain[chain_b] = chain_u
            u2b_chain[chain_u] = chain_b
            b2r_res[chain_b] = alignments[u, b][0]
            u2r_res[chain_u] = alignments[u, b][1]

    assert len(b2u_chain) == len(u2b_chain)

    # Map bound and unbound chains to common reference.
    b2r_chain = {}
    u2r_chain = {}
    reference_chain = 'A'
    for chain in b2u_chain:
        matching_chain = b2u_chain[chain]
        b2r_chain[chain] = reference_chain
        u2r_chain[matching_chain] = reference_chain
        reference_chain = chr(ord(reference_chain) + 1)

    for unmatched_u in [cu for cu in chains_u if cu not in u2b_chain]:
        logging.warning("{:} chain {:} has no match found!"
                        .format(
                            get_pdb_code_with_partner_and_binding(unbound),
                            unmatched_u))
        u2r_chain[unmatched_u] = reference_chain
        reference_chain = chr(ord(reference_chain) + 1)
    for unmatched_b in [cb for cb in chains_b if cb not in b2u_chain]:
        if style == 'dockground':
            # We don't care about unmatched bound in dockground.
            continue
        logging.warning("{:} chain {:} has no match found!"
                        .format(
                            get_pdb_code_with_partner_and_binding(bound),
                            unmatched_b))
        b2r_chain[unmatched_b] = reference_chain
        reference_chain = chr(ord(reference_chain) + 1)

    return b2r_chain, u2r_chain, b2r_res, u2r_res


def _generate_clean_unbound_bound(filename_u, filename_b, results_dir, style):
    """Perform alignment on unbound and bound files to standardize them."""
    b2r_chain, u2r_chain, b2r_res, u2r_res = \
        _get_chain_mapping(filename_u, filename_b, style)

    aug_pdb_code_u = get_pdb_code_with_partner_and_binding(filename_u)
    if style == 'db5':
        aug_pdb_code_b = get_pdb_code_with_partner_and_binding(filename_b)
        pdb_extension = 'pdb'
    elif style == 'dockground':
        partner = _get_partner(filename_u)
        aug_pdb_code_b = db.get_pdb_code(filename_b) + '_' + partner + '_b'
        pdb_extension = db.get_pdb_type(filename_b)

    output_filename_u = results_dir + '/' + aug_pdb_code_u + '_cleaned.' + \
        pdb_extension
    output_filename_b = results_dir + '/' + aug_pdb_code_b + '_cleaned.' + \
        pdb_extension
    _generate_reference(
        filename_b, b2r_chain, b2r_res, output_filename_b, style)
    _generate_reference(
        filename_u, u2r_chain, u2r_res, output_filename_u, style)

    output_mapping_u = results_dir + '/' + aug_pdb_code_u + '_toref.pkl'
    output_mapping_b = results_dir + '/' + aug_pdb_code_b + '_toref.pkl'
    _generate_mapping(u2r_chain, u2r_res, output_mapping_u)
    _generate_mapping(b2r_chain, b2r_res, output_mapping_b)


def _generate_mapping(s2r_chain, s2r_res, output_filename):
    """Write structure-to-reference chain and residue mappings."""
    mappings = {}
    mappings['chain'] = s2r_chain
    mappings['res'] = s2r_res
    with open(output_filename, 'wb') as f:
        pickle.dump(mappings, f)


def _generate_reference(pdb_filename, s2r_chain, s2r_res, output_filename,
                        style):
    """Transform PDB structure to a reference structure."""
    biopy_structure = db.parse_biopython_structure(pdb_filename)
    pdb_name = db.get_pdb_name(pdb_filename)

    new_model = Bio.PDB.Model.Model('0')
    new_structure = Bio.PDB.Structure.Structure('')
    for (chain, residues) in \
            struct.get_chain_to_valid_residues(biopy_structure, pdb_name):
        if style == 'dockground' and chain not in s2r_chain:
            # If we are in dockground, we allow ourselves to remove unmapped
            # chains.
            continue
        ref_chain = s2r_chain[chain]

        if chain in s2r_res:
            # If we have an alignment for this chain.
            new_chain = Bio.PDB.Chain.Chain(ref_chain)
            for i, residue in enumerate(residues):
                if residue.id[0] != ' ':
                    continue
                residue.segid = ""
                residue.id = (' ', s2r_res[chain][i], residue.id[2])
                new_chain.add(residue)
        else:
            # Else, just remove segment ID.
            new_chain = Bio.PDB.Chain.Chain(ref_chain)
            for i, residue in enumerate(residues):
                residue.segid = ""
        new_model.add(new_chain)

    new_structure.add(new_model)
    w = Bio.PDB.PDBIO()
    w.set_structure(new_structure)
    w.save(output_filename)


def _generate_clean_complex(pdb_code, containing_dir, results_dir, style):
    """Clean up provided complex, sending it to output_dir."""
    logging.info(pdb_code)
    lu, lb, ru, rb = get_files_for_complex(pdb_code, containing_dir, style)

    if lu is None or ru is None or lb is None or rb is None:
        logging.warning("Skipping {:} since not all pdb files present."
                        .format(pdb_code))
        return

    if not os.path.exists(results_dir):
        os.makedirs(results_dir, exist_ok=True)

#    logging.info("Receptor")
    _generate_clean_unbound_bound(ru, rb, results_dir, style)
#    logging.info("Ligand")
    _generate_clean_unbound_bound(lu, lb, results_dir, style)



"""Argument processing and routing to appropriate subcommand."""
import argparse as ap
import logging
import os
import socket
import subprocess
import sys


def add_pipeline_parsers(subparsers, pp):
    """Add new pandas and tfrecords pipeline."""
    from atom3.parse import add_parse_parser
    add_parse_parser(subparsers, pp)

    from atom3.complex import add_complexes_parser
    add_complexes_parser(subparsers, pp)

    from atom3.pair import add_pairs_parser
    add_pairs_parser(subparsers, pp)

    from atom3.conservation import add_conservation_parser
    add_conservation_parser(subparsers, pp)


def main():
    """Process all arguments."""
    p = ap.ArgumentParser(description='3D Atom Processing.')

    pp = ap.ArgumentParser(add_help=False)
    pp.add_argument('-l', metavar='log', type=str,
                    help='log file to output to (default: %(default)s)')

    subparsers = p.add_subparsers(title='subcommands',
                                  metavar='SUBCOMMAND   ',
                                  help='DESCRIPTION')

    # New pandas pipeline.
    add_pipeline_parsers(subparsers, pp)

    args = p.parse_args()

    root = logging.getLogger()
    map(root.removeHandler, root.handlers[:])
    if args.l is None:
        logging.basicConfig(stream=sys.stdout,
                            format='%(asctime)s %(levelname)s %(process)d: ' +
                            '%(message)s',
                            level=logging.INFO)
    else:
        log_dir = os.path.dirname(args.l)
        if len(log_dir) != 0 and not os.path.exists(log_dir):
            os.makedirs(log_dir, exist_ok=True)
        logging.basicConfig(filename=args.l,
                            format='%(asctime)s %(levelname)s %(process)d: ' +
                            '%(message)s',
                            level=logging.INFO)

    logging.info('=================== CALL ===================')
    logging.info('Host is {:}'.format(socket.gethostname()))
    logging.info('Git hash is {:}'.format(
        subprocess.check_output(['git', 'rev-parse', 'HEAD'])).rstrip())
    logging.info('{}'.format(' '.join(sys.argv)))
    args.func(args)
    logging.info('================= END CALL =================')


if __name__ == "__main__":
    main()



import numpy as np


# -------------------------------------------------------------------------------------------------------------------------------------
# Following code adapted from AlphaFold - PyTorch (https://github.com/Urinx/alphafold_pytorch/blob/master/feature.py):
# -------------------------------------------------------------------------------------------------------------------------------------

def extract_hmm_profile(hhm_file, sequence, asterisks_replace=0.0):
    """Extract information from an HMM file and replace asterisks. The argument hhm_file is expected to be the stringified content of the given hhm file. That is, after opening the hhm_file (i.e. open('file.hhm', 'w') as hhm_file) and calling hhm_file.read(), you should then pass these contents into this function."""
    profile_part = hhm_file.split('#')[-1]
    profile_part = profile_part.split('\n')
    whole_profile = [i.split() for i in profile_part]
    # Strip away the header and the footer
    whole_profile = whole_profile[5:-2]
    gap_profile = np.zeros((len(sequence), 10))
    aa_profile = np.zeros((len(sequence), 20))
    count_aa = 0
    count_gap = 0
    for line_values in whole_profile:
        if len(line_values) == 23:
            # The first and the last values in line_values are metadata, so skip them.
            for j, t in enumerate(line_values[2:-1]):
                aa_profile[count_aa, j] = (2 ** (-float(t) / 1000.) if t != '*' else asterisks_replace)
            count_aa += 1
        elif len(line_values) == 10:
            for j, t in enumerate(line_values):
                gap_profile[count_gap, j] = (2 ** (-float(t) / 1000.) if t != '*' else asterisks_replace)
            count_gap += 1
        elif not line_values:
            pass
        else:
            raise ValueError('Wrong length of line %s hhm file. Expected 0, 10 or 23'
                             'got %d' % (line_values, len(line_values)))
    hmm_profile = np.hstack([aa_profile, gap_profile])
    assert len(hmm_profile) == len(sequence)
    return hmm_profile


# -------------------------------------------------------------------------------------------------------------------------------------
# Following code adapted from Stack Overflow (https://stackoverflow.com/questions/4119070/how-to-divide-a-list-into-n-equal-parts-python):
# -------------------------------------------------------------------------------------------------------------------------------------
def slice_list(input_list, size):
    input_size = len(input_list)
    slice_size = input_size // size
    remain = input_size % size
    result = []
    iterator = iter(input_list)
    for i in range(size):
        result.append([])
        for j in range(slice_size):
            result[i].append(iterator.__next__())
        if remain:
            result[i].append(iterator.__next__())
            remain -= 1
    return result



"""Code to do with pdb structure sequences."""
import os
import pickle

import Bio
import Bio.PDB.Polypeptide as poly
import Bio.SeqIO
import Bio.SeqRecord
import pandas as pd

import atom3.database as db
import atom3.structure as struct


def write_fasta(seq_dict, fasta_out, id_out=None):
    """Given dictionary of name to sequence, write as FASTA."""
    records = []
    all_ids = {}
    for name, seq in seq_dict.items():
        residue_list, ids = residue_list_to_string(seq, with_ids=True)
        record = Bio.SeqRecord.SeqRecord(
            Bio.Seq.Seq(residue_list), id=name, description="")
        records.append(record)
        all_ids[name] = ids
    fasta_dir = os.path.dirname(os.path.abspath(fasta_out))
    if not os.path.exists(fasta_dir):
        os.makedirs(fasta_dir, exist_ok=True)
    Bio.SeqIO.write(records, fasta_out, "fasta")
    if id_out is not None:
        id_dir = os.path.dirname(os.path.abspath(id_out))
        if not os.path.exists(id_dir):
            os.makedirs(id_dir, exist_ok=True)
        with open(id_out, 'wb') as f:
            pickle.dump(all_ids, f)


def pdb_to_fasta(pdb_filename, fasta_filename, id_filename, separate):
    """Write a pdb file as a fasta file."""
    flat_map = {}
    pdb_name = db.get_pdb_name(pdb_filename)
    structure = pd.read_pickle(pdb_filename)
    fasta_name_to_chain = {}
    for (chain, residues) in struct.get_chain_to_valid_residues(structure):
        fasta_name = pdb_name + '-' + chain[-2] + '-' + chain[-1]
        flat_map[fasta_name] = residues
        fasta_name_to_chain[fasta_name] = chain
    names = []
    filenames = []
    id_filenames = []
    if not separate:
        write_fasta(flat_map, fasta_filename, id_out=id_filename)
        filenames.append(fasta_filename)
        id_filenames.append(id_filename)
        names.append('all')
    else:
        for (name, seq) in flat_map.items():
            new_dict = {}
            new_dict[name] = seq
            filename = fasta_filename.format(name)
            filename2 = id_filename.format(name)
            write_fasta(new_dict, filename, id_out=filename2)
            names.append(fasta_name_to_chain[name])
            filenames.append(filename)
            id_filenames.append(filename2)
    return (names, filenames, id_filenames)


def residue_list_to_string(residues, with_ids=False):
    """Convert list of residues to string."""
    for residue in residues:
        if residue.resname == 'HID':
            residue.resname = 'HIS'
        elif residue.resname == 'CYX':
            residue.resname = 'CYS'
        elif residue.resname == 'ASX':
            residue.resname = 'ASP'
        elif residue.resname == 'GLX':
            residue.resname = 'GLY'

    seq = [poly.three_to_one(residue.resname) for residue in residues
           if residue.resname != 'SEC' and residue.resname != 'PYL']
    ids = [residue.residue for residue in residues
           if residue.resname != 'SEC' and residue.resname != 'PYL']
    if with_ids:
        return "".join(seq), ids
    else:
        return "".join(seq)



"""All functions related to processing PDB datasets."""
import collections as col
import glob
import gzip
import logging
import os.path

import Bio.PDB
import numpy as np


def identity(x):
    """Return input."""
    return x


def always(x):
    """Always return true."""
    return True


def get_unique_pdbs(pdb_dataset, uniquer=identity, extension=None,
                    keeper=always):
    """
    Get pdbs in dataset that are unique according to uniquer.

    Store them in dict mapping from unique key to list of pdbs.  Keeper is
    function which determines whether we want to use the provided file.
    """
    unique_pdbs = col.defaultdict(list)
    for pdb_filename in get_structures_filenames(pdb_dataset,
                                                 extension=extension):
        if keeper(pdb_filename):
            key = uniquer(pdb_filename)
            unique_pdbs[key].append(pdb_filename)
    return unique_pdbs


def parse_biopython_structure(pdb_filename):
    """
    Extract a Bio.PDB structure from the provided PDB file.

    Args:
        pdb_filename (string):
            file path to PDB file we wish to extract Biopython structure from.

    Returns:
        Bio.PDB.Structure containing the file's data.

    Raises:
        ValueError:
            if filetype of pdb_filename not recognized.
    """
    parser = Bio.PDB.PDBParser(QUIET=True)
    _, ext = os.path.splitext(pdb_filename)

    if ext == ".gz":
        biopy_structure = parser.get_structure('pdb', gzip.open(pdb_filename, 'rt'))
    elif ".pdb" in ext:
        biopy_structure = parser.get_structure('pdb', pdb_filename)
    else:
        raise ValueError("Unrecognized filetype " + pdb_filename)
    return biopy_structure


def get_pdb_group(pdb_filename):
    """
    Get 2 letter grouping from pdb filename.

    Grab the 2 middle letters of the pdb code. E.g. 2olx pdb name gives ol as
    the group.
    """
    return get_pdb_code(pdb_filename)[1:3]


def get_pdb_basename(pdb_filename):
    """
    Get 4 letter code from pdb filename suffixed by pdb type.

    E.g. test_dir/11as_l_u_unbound.pdb1.gz has a basename of 11as.pdb1 .
    """
    return get_pdb_code(pdb_filename) + '.' + get_pdb_type(pdb_filename)


def get_pdb_type(pdb_filename):
    """
    Get type of pdb, as in extension.

    E.g. test_dir/11as.pdb1 has a type of pdb1 .
    """
    return get_pdb_name(pdb_filename).split('.')[1]


def get_pdb_code(pdb_filename):
    """
    Get 4 letter code from pdb filename.

    E.g. test_dir/11as.pdb1 has a code of 11as .
    """
    pdb_name = get_pdb_name(pdb_filename)
    if pdb_name.startswith("complex_"):
        # Haddock style name.
        return pdb_name.split('.')[0].split('_')[-1]
    else:
        return get_pdb_name(pdb_filename)[0:4]


def get_pdb_name(pdb_filename, with_type=True):
    """
    Get pdb name from full file path.

    Args:
        filename (string):
            path to pdb file.

    Returns:
        pdb name, e.g. 1aa1.pdb1
    """
    if with_type:
        return '.'.join(os.path.basename(pdb_filename).split('.')[0:2])
    else:
        return os.path.basename(pdb_filename).split('.')[0]


def get_filenames(pdb_name, pdb_dataset, extension=None, enforcement='one'):
    """
    Get path to pdb file, given pdb name and path to pdb dataset.

    Args:
        pdb_name (string):
            name of pdb file, e.g. 1aa1.pdb1
        pdb_dataset (string):
            path to pdb dataset.

    Returns:
        path to file.  If multiple matches, return first one.
    """
    filenames = get_structures_filenames(pdb_dataset, extension=extension)
    good = [x for x in filenames if pdb_name in x]
    _check_filenames(pdb_name, good, enforcement=enforcement)
    return good


def get_all_filenames(pdb_names, pdb_dataset, extension=None,
                      keyer=get_pdb_name, enforcement='one'):
    """Return filenames for provided pdb names, in same order."""
    filenames = get_structures_filenames(pdb_dataset, extension=extension)
    dict = col.defaultdict(list)
    for filename in filenames:
        dict[keyer(filename)].append(filename)
    res = []
    for pdb_name in pdb_names:
        good = dict[pdb_name]
        _check_filenames(pdb_name, good, enforcement=enforcement)
        res.append(good)
    return res


def _check_filenames(pdb_name, pdb_filenames, enforcement='one'):
    """Make sure we have the right number of filenames."""
    if enforcement == 'one' and len(pdb_filenames) != 1:
        logging.error(
            "Need exactly one file per key, instead for {:} got {:} ({:})"
                .format(pdb_name, len(pdb_filenames), pdb_filenames))
        raise RuntimeError
    elif enforcement == 'at_least_one' and len(pdb_filenames) < 1:
        logging.error(
            "Need at least one file per key, instead for {:} got {:} "
            "({:})"
                .format(pdb_name, len(pdb_filenames), pdb_filenames))
        raise RuntimeError
    elif enforcement == 'none':
        pass


def get_structures_filenames(pdb_dataset, extension=None):
    """
    Return structure filenames from PDB database specified in bio_dir file.

    Args:
        bio_dir (string):
            location of PDB snapshot's biounit divided directory (i.e.
            snapshot/data/biounit/PDB/divided)
    """
    # Cache snapshot filenames here since a bit expensive.

    if _is_pdb_list(pdb_dataset):
        return pdb_dataset

    if _is_pdb_regex(pdb_dataset):
        return glob.glob(pdb_dataset)

    if _is_pdb_file(pdb_dataset, extension=extension):
        return [pdb_dataset]

    if _is_pdb_list_in_file(pdb_dataset, extension=extension):
        return _get_pdb_list_from_file(pdb_dataset, extension=extension)

    if _is_pdb_dir(pdb_dataset, extension=extension):
        return _get_pdb_dir_filenames(pdb_dataset, extension=extension)

    pdb_snapshot_filenames = _get_pdb_snapshot_filenames(
        pdb_dataset, extension=extension)
    if len(pdb_snapshot_filenames) != 0:
        return pdb_snapshot_filenames

    return []


def _is_pdb_list_in_file(pdb_file_list, extension=None):
    """Check if provided file contains a list of pdbs."""
    if not os.path.isfile(pdb_file_list):
        return False
    return len(_get_pdb_list_from_file(pdb_file_list, extension)) != 0


def _is_pdb_dir(pdb_dir, recurse=True, extension=None):
    """Check if provided file is a directory containing pdbs."""
    if not os.path.isdir(pdb_dir):
        return False
    return len(_get_pdb_dir_filenames(pdb_dir, recurse=recurse,
                                      extension=extension)) != 0


def _is_pdb_list(pdb_list):
    """Check if provided dataset is list of pdb files."""
    if isinstance(pdb_list, list) or isinstance(pdb_list, np.ndarray):
        return True
    return False


def _is_pdb_file(pdb_file, extension=None):
    """Check if provided file is a pdb file."""
    if not os.path.isfile(pdb_file):
        return False
    _, ext = os.path.splitext(pdb_file)
    if extension is None:
        return ext == ".gz" or ".pdb" in ext
    else:
        return ext == extension


def _is_pdb_regex(pdb_regex):
    """Check if provided file is a pdb regex."""
    return '*' in pdb_regex


def _is_pdb_snapshot(pdb_snapshot, extension=None):
    """Check if provided file is a directory of pdb groups."""
    return len(_get_pdb_snapshot_filenames(
        pdb_snapshot, extension=extension)) != 0


def _get_pdb_list_from_file(pdb_file_list, extension=None):
    """Get pdb files from a text file containing one pdb file per line."""
    with open(pdb_file_list, 'r') as f:
        content = f.readlines()
    candidate_files = [x.rstrip() for x in content]
    pdb_files = [x for x in candidate_files
                 if _is_pdb_file(x, extension=extension)]
    return pdb_files


def _get_pdb_dir_filenames(pdb_dir, recurse=True, extension=None):
    """Get files, given a pdb dir path."""
    pdb_filenames = []

    if recurse:
        for file in os.listdir(pdb_dir):
            new_dir = os.path.join(pdb_dir, file)
            if os.path.isdir(new_dir):
                pdb_filenames += _get_pdb_dir_filenames(new_dir, recurse=False,
                                                        extension=extension)

    if extension is None:
        newly_discovered_pdb_filenames = glob.glob(pdb_dir + '/*.pdb*.gz') + \
                                         glob.glob(pdb_dir + '/*.pdb[0-9]') + \
                                         glob.glob(pdb_dir + '/*.pdb')
        pdb_filenames += newly_discovered_pdb_filenames
    else:
        newly_discovered_pdb_filenames = glob.glob(pdb_dir + '/*' + extension)
        pdb_filenames += newly_discovered_pdb_filenames

    pdb_filenames = [x for x in pdb_filenames if os.path.isfile(x)]

    return pdb_filenames


def _get_pdb_snapshot_filenames(pdb_snapshot, extension=None):
    """Get files, given a pdb snapshot path."""
    pdb_groups = glob.glob(pdb_snapshot + '/[A-Za-z0-9][A-Za-z0-9]')
    valid = False
    for group in pdb_groups:
        if _is_pdb_dir(group, recurse=False, extension=extension):
            valid = True
            break

    if not valid:
        # Try and see if it stored under RCSB snapshot location.
        pdb_snapshot = pdb_snapshot + '/data/biounit/PDB/divided'
        pdb_groups = glob.glob(pdb_snapshot + '/[a-z0-9][a-z0-9]')

    pdb_filenames = []
    for pdb_group in pdb_groups:
        pdb_filenames.extend(_get_pdb_dir_filenames(pdb_group, recurse=False,
                                                    extension=extension))
    return pdb_filenames



"""
BioPython PDB Sructure wrapper class.

Allows for extraction of regions from the structure.
"""
import logging
import os
import re

import Bio.PDB
import Bio.PDB.Polypeptide as poly
import numpy as np
import pandas as pd

import atom3.database as db

# Max number of residues allowed.  Prevents processing of massive PDB
# structures.
max_residues = 100000

expected = {
    "TRP": 14,
    "PHE": 11,
    "LYS": 9,
    "PRO": 7,
    "ASP": 8,
    "ALA": 5,
    "ARG": 11,
    "CYS": 6,
    "VAL": 7,
    "THR": 7,
    "GLY": 4,
    "SER": 6,
    "HIS": 10,
    "LEU": 8,
    "GLU": 9,
    "TYR": 12,
    "ILE": 8,
    "ASN": 8,
    "MET": 8,
    "GLN": 9,
}


def extract_c_alpha_regions(structure, radius_ang, detailed=False):
    """
    Given a PDB filename, extracts biopython structure and then its regions.

    The regions are centered at the residue c-alphas.

    Args:
        pdb_filename (string):
            path to PDB file.
        radius_ang (int):
            radius of the surfacelets, in angstroms.

    Returns:
        Structured array of src.feat.region the regions extracted.
    """
    return structure.generate_regions(
        radius_ang,
        structure.generate_c_alpha_centroids(),
        detailed=detailed)


def parse_structure(structure_filename, concoord=False, one_model=False):
    """Parse a file into chain,model-to-residue mapping."""
    _, ext = os.path.splitext(structure_filename)
    detailed = ext == '.pkl'
    if detailed:
        # If detailed we are reading pandas pickle file outputted by
        # protprep.
        df = pd.read_pickle(structure_filename)
        # Set model to 0, because a multi-model file was either already split
        # into separate files (using the split command) or was pared down to a
        # single model by the autodock portion of the protprep pipeline.
        # This might need to be revisited if/when autodock is removed from
        # pipeline or we decide to actually keep track of correct model.
        df['model'] = get_model(structure_filename)
        # Remove hydrogens, for now, to maintain compatability.
        df = df[df['maestro_atom_name'].apply(lambda x: x.strip()[0]) != 'H']
    else:
        # BioPython.PDB Structure extracted from PDB file.
        biopy_structure = db.parse_biopython_structure(structure_filename)
        pdb_name = db.get_pdb_name(structure_filename)
        if concoord:
            # need to set model number to be correct (drawn from filename)
            # TODO: I (Raphael) moved this out of core Structure code, need to
            # make sure it is correct still for CONCOORD.
            biopy_structure = db.parse_biopython_structure(structure_filename)
            biopy_structure = \
                Bio.PDB.Structure.Structure(biopy_structure.id)

            chainmodel = pdb_name.split('_')[1]
            model_id = str(int(re.split('(\d+)', chainmodel)[1]) + 1)

            for model_obj in biopy_structure:
                new_model = Bio.PDB.Model.Model(model_id)
                for chain in model_obj:
                    new_model.add(chain)
                biopy_structure.add(new_model)

        if one_model:
            new_structure = Bio.PDB.Structure.Structure(biopy_structure.id)
            new_structure.add(biopy_structure[0])
            biopy_structure = new_structure

        atoms = []
        for residue in Bio.PDB.Selection.unfold_entities(biopy_structure, 'R'):
            # Prune out things that aren't actually residue atoms.
            if 'CA' in residue and residue.get_id()[0] == ' ':
                for atom in residue:
                    atoms.append(atom)

        df = pd.DataFrame([(
            pdb_name,
            str(atom.get_parent().get_parent().get_parent().serial_num),
            atom.get_parent().get_full_id()[2],
            str(atom.get_parent().get_id()[1]) + atom.get_parent().get_id()[2],
            atom.get_parent().get_resname(),
            atom.get_coord()[0],
            atom.get_coord()[1],
            atom.get_coord()[2],
            atom.get_id()[0],
            atom.get_name(),
            str(atom.serial_number)) for atom in atoms],
            columns=[
                'pdb_name',
                'model',
                'chain',
                'residue',
                'resname',
                'x',
                'y',
                'z',
                'element',
                'atom_name',
                'aid'])
    return df


def get_ca_pos_from_residues(df, res):
    """Look up alpha carbon positions of provided residues."""
    ca = df[df['atom_name'] == 'CA'].reset_index().set_index(
        ['pdb_name', 'model', 'chain', 'residue'])
    nb = ca.reindex(res)
    nb = nb.reset_index().set_index('index')
    return nb


def get_ca_pos_from_atoms(df, atoms):
    """Look up alpha carbon positions of provided atoms."""
    ca = df[df['atom_name'] == 'CA'].reset_index()
    nb = ca.reindex(atoms)
    nb = nb.reset_index().set_index('index')
    return nb


def get_chain_to_valid_residues(structure, pdb_name=None):
    """Get tuples of chains and their valid residues."""
    if pdb_name is None:
        pdb_name = ()
    else:
        pdb_name = (pdb_name,)
    chain_res = []
    if type(structure) is Bio.PDB.Structure.Structure:
        for model in structure:
            for chain in model:
                residues = [res for res in chain
                            if poly.is_aa(res.get_resname(), standard=True) and
                            'CA' in res]
                if len(residues) != 0:
                    chain_res.append((
                        pdb_name + (str(model.serial_num), chain.get_id()),
                        residues))
    else:
        if 'atom_name' in structure.columns:
            calphas = structure[structure['atom_name'] == 'CA']
        else:
            calphas = structure[structure['maestro_atom_name'] == ' CA ']
        calphas = calphas[calphas['resname'] != 'UNK']

        for (chain, chain_ca) in calphas.groupby(['model', 'chain']):
            residues = [ca for idx, ca in chain_ca.iterrows()]
            if len(residues) != 0:
                chain_res.append((pdb_name + chain, residues))
    return chain_res


def df_to_ca(df):
    """Extract only alpha carbons from dataframe."""
    ca = df[df['atom_name'] == 'CA']
    ca.index.name = 'df_index'
    ca = ca.reset_index()
    return ca


def create_lookup(ca):
    lookup = {}

    def lookup_aa_id(pdb_name, model, chain, residue):
        """Lookup amino acid id based on pdb_name, model, chain, residue id."""
        if (pdb_name, model, chain, residue) not in lookup:
            res = ca.loc[
                (ca['residue'] == residue) &
                (ca['chain'] == chain) &
                (ca['model'] == model) &
                (ca['pdb_name'] == pdb_name)].index
            assert len(res) == 1
            lookup[(pdb_name, model, chain, residue)] = res[0]

        return lookup[(pdb_name, model, chain, residue)]

    return lookup_aa_id


def add_sidechains_parser(subparsers, pp):
    """Add parser."""

    def get_missing_sidechains_main(args):
        get_missing_sidechains(args.pdb_dataset, args.output_scwrl)

    ap = subparsers.add_parser(
        'sidechains', description="find missing side chains",
        help='generates a sequence for use with scwrl4 where missing side'
             ' residues are upper-cased',
        parents=[pp])
    ap.set_defaults(func=get_missing_sidechains_main)
    ap.add_argument('pdb_dataset', type=str,
                    help='pdb dataset to get missing side chains for.')
    ap.add_argument('output_scwrl', type=str,
                    help='where to write output sequences to.')


def get_missing_sidechains(pdb_dataset, output_scwrl):
    """Get residues that are missing atoms."""
    for pdb_filename in db.get_structures_filenames(pdb_dataset):
        biopy_structure = db.parse_biopython_structure(pdb_filename)
        pdb_name = db.get_pdb_name(pdb_filename)
        missing = 0
        scwrl_list = []
        logging.info("Processing {:}".format(pdb_name))
        for model in biopy_structure:
            for chain in model:
                for i, residue in enumerate(chain):
                    res_name = residue.resname
                    if res_name not in expected:
                        logging.warning("Non-standard residue found: {:}. "
                                        "Skipping.".format(res_name))
                        continue
                    res_code = poly.three_to_one(res_name)
                    res_id = residue.id[1]
                    curr_count = len(
                        Bio.PDB.Selection.unfold_entities(residue, 'A'))
                    if curr_count != expected[res_name]:
                        logging.debug(
                            "Missing residue {:} at position {:} (with id {:})"
                            " which has {:} instead of the expected {:} atoms."
                                .format(res_name, i, res_id, curr_count,
                                        expected[res_name]))
                        missing += 1
                        scwrl_list.append(res_code.upper())
                    else:
                        scwrl_list.append(res_code.lower())

        logging.debug("Missing {:} residue total".format(missing))
        with open(output_scwrl, 'w') as f:
            f.write("".join(scwrl_list))


def df_to_sarray(df):
    """
    Convert a pandas DataFrame object to a numpy structured array.
    This is functionally equivalent to but more efficient than
    np.array(df.to_array())

    :param df: the data frame to convert
    :return: a numpy structured array representation of df
    """

    v = df.values
    cols = df.columns

    types = [(cols[i].encode(), df[k].dtype.type)
             for (i, k) in enumerate(cols)]
    dtype = np.dtype(types)
    z = np.zeros(v.shape[0], dtype)
    for (i, k) in enumerate(z.dtype.names):
        z[k] = v[:, i]
    return z


def get_model(pdb_filename):
    """Get model from split pdb filename."""
    pdb_name = db.get_pdb_name(pdb_filename, with_type=False)
    tokens = pdb_name.split("_")
    if len(tokens) < 3:
        return 0
    elif not tokens[2].isdigit():
        return 0
    else:
        return int(tokens[2])


def get_chain(pdb_filename):
    """Get chain from split pdb filename."""
    pdb_name = db.get_pdb_name(pdb_filename, with_type=False)
    tokens = pdb_name.split("_")
    if len(tokens) < 3:
        return 0
    else:
        return tokens[1]


def df_to_pdb(df_in):
    """Convert df to pdb."""
    df = df_in.copy()
    new_structure = Bio.PDB.Structure.Structure('')
    for (model, m_atoms) in df.groupby(['model']):
        new_model = Bio.PDB.Model.Model(model)
        for (chain, c_atoms) in m_atoms.groupby(['chain']):
            new_chain = Bio.PDB.Chain.Chain(chain)
            for (residue, r_atoms) in c_atoms.groupby(['residue']):
                new_residue = residue_to_pdbresidue(residue, r_atoms)
                new_chain.add(new_residue)
            new_model.add(new_chain)
        new_structure.add(new_model)
    return new_structure


def residue_to_pdbresidue(residue, r_atoms):
    resname = r_atoms['resname'].unique()
    assert len(resname) == 1
    resname = resname[0]
    new_residue = Bio.PDB.Residue.Residue(
        (' ', int(residue[:-1]), residue[-1]), resname, '')
    for row, atom in r_atoms.iterrows():
        new_atom = Bio.PDB.Atom.Atom(
            atom['atom_name'],
            [atom['x'], atom['y'], atom['z']],
            1,
            1,
            ' ',
            atom['atom_name'],
            atom['aid'],
            atom['element'])
        new_residue.add(new_atom)
    return new_residue



from setuptools import setup

setup(
    name='atom3-py3',
    packages=['atom3'],
    url='https://github.com/amorehead/atom3',
    version='0.1.8.3',
    description='3D Atomic Data Processing (Python 3 version)',
    long_description=open("README.rst").read(),
    author='Raphael Townshend',
    license='MIT',
    install_requires=[
        'biopython',
        'click',
        'easy-parallel-py3',
        'h5py',
        'pandas>=0.24.0',
        'scipy',
    ],
    entry_points={
        'console_scripts': [
            "atom3 = atom3.main:main",
        ]
    }
)



"""Functions relating to pairs of regions."""

import numpy as np
import pandas as pd
import scipy.spatial as spa


def build_get_neighbors(criteria, cutoff):
    """Build a neighbor function to check if two subunits are neighboring."""

    def get_neighbors_param(df0, df1):
        return get_neighbors(df0, df1, criteria, cutoff)

    return get_neighbors_param


def get_neighbors(df0, df1, criteria, cutoff):
    """Find neighbors according to provided criteria and cutoff."""
    if criteria == 'non_heavy_res':
        return _get_non_heavy_neighbor_residues(df0, df1, cutoff)
    elif criteria == 'non_heavy_atom':
        return _get_non_heavy_neighbor_atoms(df0, df1, cutoff)
    elif criteria == 'ca_res':
        return _get_ca_neighbor_residues(df0, df1, cutoff)
    elif criteria == 'ca_atom':
        return _get_ca_neighbor_atoms(df0, df1, cutoff)
    else:
        raise RuntimeError("Unrecognized criteria {:}".format(criteria))


def _get_non_heavy_neighbor_residues(df0, df1, cutoff):
    """Get neighboring residues for non-heavy atom-based distance."""
    non_heavy0 = df0[df0['element'] != 'H']
    non_heavy1 = df1[df1['element'] != 'H']

    dist = spa.distance.cdist(non_heavy0[['x', 'y', 'z']], non_heavy1[['x', 'y', 'z']])
    pairs = np.array(np.where(dist < cutoff)).T
    if len(pairs) == 0:
        return [], []
    # Use the found pairs to find unique pairings of residues.
    res0 = non_heavy0.iloc[pairs[:, 0]][['pdb_name', 'model', 'chain', 'residue']]
    res1 = non_heavy1.iloc[pairs[:, 1]][['pdb_name', 'model', 'chain', 'residue']]
    res0 = res0.reset_index(drop=True)
    res1 = res1.reset_index(drop=True)
    # We concatenate so that we can find unique _pairs_.
    res = pd.concat((res0, res1), axis=1)
    res = res.drop_duplicates()
    # # Split back out now that we have found duplicates.
    res0 = res.iloc[:, range(4)]
    res1 = res.iloc[:, range(4, 8)]
    res0 = res0.reset_index(drop=True)
    res1 = res1.reset_index(drop=True)

    return res0, res1


def _get_non_heavy_neighbor_atoms(df0, df1, cutoff):
    """Get neighboring atoms for non-heavy atom-based distance."""
    non_heavy0 = df0[df0['element'] != 'H']
    non_heavy1 = df1[df1['element'] != 'H']

    dist = spa.distance.cdist(non_heavy0[['x', 'y', 'z']], non_heavy1[['x', 'y', 'z']])
    pairs = np.array(np.where(dist < cutoff)).T
    if len(pairs) == 0:
        return [], []
    # Use the found pairs to find unique pairings of atoms.
    atoms0 = non_heavy0.iloc[pairs[:, 0]]
    atoms1 = non_heavy1.iloc[pairs[:, 1]]
    atoms0 = atoms0.reset_index(drop=True)
    atoms1 = atoms1.reset_index(drop=True)
    # Concatenate so that we can find unique pairs.
    atoms = pd.concat((atoms0, atoms1), axis=1)
    atoms = atoms.drop_duplicates()
    # Split back out now that we have found duplicates.
    atoms0 = atoms.iloc[:, range(11)]
    atoms1 = atoms.iloc[:, range(11, 22)]
    atoms0 = atoms0.reset_index(drop=True)
    atoms1 = atoms1.reset_index(drop=True)
    return atoms0, atoms1


def _get_ca_neighbor_residues(df0, df1, cutoff):
    """Get neighboring residues for alpha-carbon based distance."""
    ca0 = df0[df0['atom_name'] == 'CA']
    ca1 = df1[df1['atom_name'] == 'CA']

    dist = spa.distance.cdist(ca0[['x', 'y', 'z']], ca1[['x', 'y', 'z']])
    pairs = np.array(np.where(dist < cutoff)).T
    if len(pairs) == 0:
        return [], []
    res0 = ca0.iloc[pairs[:, 0]][['pdb_name', 'model', 'chain', 'residue']]
    res1 = ca1.iloc[pairs[:, 1]][['pdb_name', 'model', 'chain', 'residue']]
    res0 = res0.reset_index(drop=True)
    res1 = res1.reset_index(drop=True)
    return res0, res1


def _get_ca_neighbor_atoms(df0, df1, cutoff):
    """Get neighboring atoms for alpha-carbon based distance."""
    ca0 = df0[df0['atom_name'] == 'CA']
    ca1 = df1[df1['atom_name'] == 'CA']

    dist = spa.distance.cdist(ca0[['x', 'y', 'z']], ca1[['x', 'y', 'z']])
    pairs = np.array(np.where(dist < cutoff)).T
    if len(pairs) == 0:
        return [], []
    atoms0 = ca0.iloc[pairs[:, 0]]
    atoms1 = ca1.iloc[pairs[:, 1]]
    atoms0 = atoms0.reset_index(drop=True)
    atoms1 = atoms1.reset_index(drop=True)
    return atoms0, atoms1



import collections as col
import logging
import multiprocessing as mp
import os

import dill
import numpy as np
import pandas as pd
import parallel as par

import atom3.case as ca
import atom3.complex as comp
import atom3.database as db
import atom3.neighbors as nb
from atom3.structure import get_ca_pos_from_residues, get_ca_pos_from_atoms

Pair = col.namedtuple(
    'Pair', ['complex', 'df0', 'df1', 'pos_idx', 'neg_idx', 'srcs', 'id', 'sequences'])

sem = mp.Semaphore()


def add_pairs_parser(subparsers, pp):
    """Add parser."""

    def all_complexes_to_pairs_main(args):
        all_complexes_to_pairs_full(args)

    ap = subparsers.add_parser(
        'pairs', description='complexes to pairs',
        help='uses output of the complex command to split pickled proteins',
        parents=[pp])
    ap.set_defaults(func=all_complexes_to_pairs_main)
    ap.add_argument('nbd', metavar='neighbor_def', type=str,
                    help='neighbor definition (i.e. non_heavy_res, non_heavy_atom, ca_res, or ca_atom)')
    ap.add_argument('complexes_dill', metavar='complexes.dill', type=str,
                    help='complexes file')
    ap.add_argument('output_dir', type=str,
                    help='directory to output to')
    ap.add_argument('-n', '--criteria', dest='criteria',
                    choices=['ca', 'heavy'],
                    default='ca', help='criteria for finding neighboring'
                                       ' residues (default: by alpha carbon distance)')
    ap.add_argument('-t', '--cutoff', dest='cutoff', type=float,
                    default=8, help='cutoff distance to be used with'
                                    ' neighbor criteria (default: 8)')
    ap.add_argument('-f', '--full', dest='full', action='store_true',
                    help='generate all possible negative examples, '
                         'as opposed to a sampling of same size as positive.')
    ap.add_argument(
        '-u', '--unbound', help='whether to use unbound data.',
        action="store_true")
    ap.add_argument('-c', metavar='cpus', default=mp.cpu_count(), type=int,
                    help='number of cpus to use for processing (default:'
                         ' number processors available on current machine)')


def all_complexes_to_pairs_full(args):
    complexes = comp.read_complexes(args.complexes_dill)
    get_neighbors = nb.build_get_neighbors(args.criteria, args.cutoff)
    get_pairs = build_get_pairs(args.nbd, complexes['type'], args.unbound, get_neighbors, args.full)
    all_complex_to_pairs(complexes, get_pairs, args.output_dir, args.c)


def all_complex_to_pairs(complexes, get_pairs, output_dir, num_cpus):
    """Reads in structures and produces appropriate pairings."""
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)

    requested_keys = complexes['data'].keys()
    produced_keys = complexes_from_pair_dir(output_dir)
    work_keys = [key for key in requested_keys if key not in produced_keys]

    inputs = [(complexes['data'][key], get_pairs, output_dir)
              for key in work_keys]
    logging.info("{:} requested keys, {:} produced keys, {:} work keys"
                 .format(len(requested_keys), len(produced_keys),
                         len(work_keys)))
    par.submit_jobs(complex_to_pairs, inputs, num_cpus)


def complexes_from_pair_dir(pair_dir):
    """Get all complex names from provided pair directory."""
    filenames = db.get_structures_filenames(pair_dir, extension='.dill')
    # Remove per-chain identifier.
    # TODO: This could cause issues when only some of the pairs have been
    # written.
    return ['_'.join(db.get_pdb_name(x).split('_')[:-1]) for x in filenames]


def complex_to_pairs(complex, get_pairs, output_dir):
    pairs_txt = output_dir + '/pairs.txt'
    name = complex.name
    logging.info("Working on {:}".format(name))
    pairs, num_subunits = get_pairs(complex)
    logging.info("For complex {:} found {:} pairs out of {:} chains"
                 .format(name, len(pairs), num_subunits))
    sub_dir = output_dir + '/' + db.get_pdb_code(name)[1:3]
    f = name
    if ('mut' in f) and ('mut' not in db.get_pdb_code(name)):
        pdb = db.get_pdb_code(name)(f) + f[f.rfind('_') + 1: f.find('.')]
        sub_dir = output_dir + '/' + pdb
    with sem:
        if len(pairs) > 0:
            if not os.path.exists(sub_dir):
                os.makedirs(sub_dir, exist_ok=True)
            with open(pairs_txt, 'a') as f:
                f.write(name + '\n')

    for i, pair in enumerate(pairs):
        output_dill = "{:}/{:}_{:}.dill".format(sub_dir, name, i)
        write_pair_as_dill(pair, output_dill)


def write_pair_as_dill(pair, output_dill):
    """Write pair as dill file."""
    with open(output_dill, 'wb') as f:
        dill.dump(pair, f)


def read_pair_from_dill(input_dill):
    """Read pair from dill file."""
    with open(input_dill, 'rb') as f:
        return dill.load(f)


def build_get_pairs(neighbor_def, type, unbound, nb_fn, full):
    def get_pairs_param(complex):
        return get_pairs(neighbor_def, complex, type, unbound, nb_fn, full)

    return get_pairs_param


def get_pairs(neighbor_def, complex, type, unbound, nb_fn, full):
    """
    Get pairings for provided complex.

    A complex is a set of chains.  For our interface prediction problem, we
    currently only deal with pairs of chains.  Here, we find all possible such
    pairings, for a given definition of neighboring.
    """
    if type == 'rcsb':
        pairs, num_subunits = \
            _get_rcsb_pairs(neighbor_def, complex, unbound, nb_fn, full)
    elif type == 'db5' or type == 'db5mut' or type == 'hotspot':
        pairs, num_subunits = \
            _get_db5_pairs(complex, unbound, nb_fn, full)
    elif type == 'dockground':
        pairs, num_subunits = \
            _get_db5_pairs(complex, unbound, nb_fn, full)
    else:
        raise RuntimeError("Unrecognized dataset type {:}".format(type))
    return pairs, num_subunits


def _get_rcsb_pairs(neighbor_def, complex, unbound, nb_fn, full):
    """
    Get pairs for rcsb type complex.

    For this type of complex, we assume that each chain is its own entity,
   and that two chains form a pair if at least one pair of residues spanning
    the two are considered neighbors.
    """
    if unbound:
        logging.error("Requested unbound pairs from RCSB type complex, "
                      "even though they don't have unbound data.")
        raise RuntimeError("Unbound requested for RCSB")
    (pkl_filename,) = complex.bound_filenames
    df = pd.read_pickle(pkl_filename)
    # TODO: Allow for keeping more than just first model.
    if df.shape[0] == 0:
        return [], 0
    df = df[df['model'] == df['model'][0]]
    pairs, num_chains = _get_all_chain_pairs(neighbor_def, complex, df, nb_fn, pkl_filename, full)
    return pairs, num_chains


def _get_db5_pairs(complex, unbound, nb_fn, full):
    """
    Get pairs for docking benchmark 5 type complex.
    For this type of complex, we assume that each file is its own entity,
    and that there is essentially one pair for each complex, with one side
    being all the chains of the ligand, and the other all the chains of the
    receptor.
    """
    (lb, rb) = complex.bound_filenames
    (lu, ru) = complex.unbound_filenames
    lb_df = pd.read_pickle(lb)
    rb_df = pd.read_pickle(rb)
    # Always use bound to get neighbors...
    lres, rres = nb_fn(lb_df, rb_df)
    if unbound:
        # ...but if unbound, we then use the actual atoms from unbound.
        ldf, rdf = pd.read_pickle(lu), pd.read_pickle(ru)

        # Convert residues' pdb_names to unbound.
        lres['pdb_name'] = lres['pdb_name'].map(
            lambda x: ca.find_of_type(x, ldf['pdb_name'].values, None, False, style='db5'))
        rres['pdb_name'] = rres['pdb_name'].map(
            lambda x: ca.find_of_type(x, rdf['pdb_name'].values, None, False, style='db5'))

        # Remove residues that we cannot map from bound structure to unbound.
        rres_index = rres[['pdb_name', 'model', 'chain', 'residue']]
        lres_index = lres[['pdb_name', 'model', 'chain', 'residue']]
        rdf_index = rdf[['pdb_name', 'model', 'chain', 'residue']]
        ldf_index = ldf[['pdb_name', 'model', 'chain', 'residue']]
        rgone = [i for i, x in rres_index.iterrows()
                 if not (np.array(x) == rdf_index).all(1).any()]
        lgone = [i for i, x in lres_index.iterrows()
                 if not (np.array(x) == ldf_index).all(1).any()]
        gone = list(set(lgone).union(set(rgone)))
        if len(gone) > 0:
            logging.warning("Dropping {:}/{:} residues from {:} that didn't map "
                            "to unbound from bound.".format(len(gone), len(lres), complex.name))
            lres = lres.drop(gone)
            rres = rres.drop(gone)

        lsrc, rsrc = lu, ru
    else:
        ldf, rdf = lb_df, rb_df
        lsrc, rsrc = lb, rb
    lpos = get_ca_pos_from_residues(ldf, lres)
    rpos = get_ca_pos_from_residues(rdf, rres)
    pos_idx, neg_idx = _get_residue_positions(ldf, lpos, rdf, rpos, full)
    srcs = {'src0': lsrc, 'src1': rsrc}
    pair = Pair(complex=complex.name, df0=ldf, df1=rdf, pos_idx=pos_idx, neg_idx=neg_idx, srcs=srcs, id=0, sequences={})
    return [pair], 2


def _get_all_chain_pairs(neighbor_def, complex, df, nb_fn, filename, full):
    """Get all possible chain pairs from provided dataframe."""
    pairs = []
    # We reset the index here so each chain's dataframe can be treated independently.
    groups = [(x[0], x[1].reset_index(drop=True))
              for x in df.groupby(['chain', 'model'])]
    num_chains = len(groups)
    num_pairs = 0
    pair_idx = 0
    for i in range(num_chains):
        (chain0, df0) = groups[i]
        for j in range(i + 1, num_chains):
            (chain1, df1) = groups[j]
            if 'atom' in neighbor_def:
                atoms0, atoms1 = nb_fn(df0, df1)
                if len(atoms0) == 0:
                    # No neighbors between these 2 chains.
                    continue
                else:
                    num_pairs += 1
                pos0 = get_ca_pos_from_atoms(df0, atoms0)
                pos1 = get_ca_pos_from_atoms(df1, atoms1)
                pos_idx, neg_idx = _get_atoms_positions(df0, pos0, df1, pos1, full)
            else:  # Use residue-level neighbor function
                res0, res1 = nb_fn(df0, df1)
                if len(res0) == 0:
                    # No neighbors between these 2 chains.
                    continue
                else:
                    num_pairs += 1
                pos0 = get_ca_pos_from_residues(df0, res0)
                pos1 = get_ca_pos_from_residues(df1, res1)
                pos_idx, neg_idx = _get_residue_positions(df0, pos0, df1, pos1, full)
            srcs = {'src0': filename, 'src1': filename}
            pair = Pair(complex=complex.name, df0=df0, df1=df1, pos_idx=pos_idx,
                        neg_idx=neg_idx, srcs=srcs, id=pair_idx, sequences={})
            pairs.append(pair)
            pair_idx += 1
    return pairs, num_chains


def _get_residue_positions(df0, pos_ca0, df1, pos_ca1, full):
    """Get negative pairings given positive pairings."""
    ca0 = df0[df0['atom_name'] == 'CA']
    ca1 = df1[df1['atom_name'] == 'CA']
    num0, num1 = ca0.shape[0], ca1.shape[0]
    num_pos = pos_ca0.shape[0]
    num_total = num0 * num1
    pos_idxs = []
    for p0, p1 in zip(pos_ca0.index, pos_ca1.index):
        idx0 = ca0.index.get_loc(p0)
        idx1 = ca1.index.get_loc(p1)
        pos_idxs.append((idx0, idx1))
    pos_idxs = np.array(pos_idxs)
    pos_flat = np.ravel_multi_index(
        (pos_idxs[:, 0], pos_idxs[:, 1]), (num0, num1))
    neg_flat = np.arange(num_total)
    neg_flat = np.delete(neg_flat, pos_flat)
    if not full:
        np.random.shuffle(neg_flat)
        neg_flat = neg_flat[:num_pos]
    neg_idxs = np.array(np.unravel_index(neg_flat, (num0, num1))).T
    neg_ca0 = ca0.iloc[neg_idxs[:, 0]]
    neg_ca1 = ca1.iloc[neg_idxs[:, 1]]
    neg_ca_idxs = np.stack((neg_ca0.index.values, neg_ca1.index.values)).T
    pos_ca_idxs = np.stack((pos_ca0.index.values, pos_ca1.index.values)).T
    return pos_ca_idxs, neg_ca_idxs


def _get_atoms_positions(df0, pos_ca0, df1, pos_ca1, full):
    """Get negative pairings given positive pairings."""
    ca0 = df0[df0['atom_name'] == 'CA']
    ca1 = df1[df1['atom_name'] == 'CA']
    num0, num1 = ca0.shape[0], ca1.shape[0]
    num_pos = pos_ca0.shape[0]
    num_total = num0 * num1
    pos_idxs = []
    for p0, p1 in zip(pos_ca0.index, pos_ca1.index):
        idx0 = ca0.index.get_loc(p0)
        idx1 = ca1.index.get_loc(p1)
        pos_idxs.append((idx0, idx1))
    pos_idxs = np.array(pos_idxs)
    pos_flat = np.ravel_multi_index(
        (pos_idxs[:, 0], pos_idxs[:, 1]), (num0, num1))
    neg_flat = np.arange(num_total)
    neg_flat = np.delete(neg_flat, pos_flat)
    if not full:
        np.random.shuffle(neg_flat)
        neg_flat = neg_flat[:num_pos]
    neg_idxs = np.array(np.unravel_index(neg_flat, (num0, num1))).T
    neg_ca0 = ca0.iloc[neg_idxs[:, 0]]
    neg_ca1 = ca1.iloc[neg_idxs[:, 1]]
    neg_ca_idxs = np.stack((neg_ca0.index.values, neg_ca1.index.values)).T
    pos_ca_idxs = np.stack((pos_ca0.index.values, pos_ca1.index.values)).T
    return pos_ca_idxs, neg_ca_idxs


def _get_atom_positions(df0, pos_atoms0, df1, pos_atoms1, full):
    """Get negative atom pairings given positive atom pairings."""
    non_heavy0 = df0[df0['element'] != 'H']
    non_heavy1 = df1[df1['element'] != 'H']
    num0, num1 = non_heavy0.shape[0], non_heavy1.shape[0]
    num_pos = pos_atoms0.shape[0]
    num_total = num0 * num1
    pos_idxs = []
    for p0, p1 in zip(pos_atoms0.index, pos_atoms1.index):
        idx0 = non_heavy0.index.get_loc(p0)
        idx1 = non_heavy1.index.get_loc(p1)
        pos_idxs.append((idx0, idx1))
    pos_idxs = np.array(pos_idxs)
    pos_flat = np.array([0]) if pos_idxs.size == 0 else np.ravel_multi_index(
        (pos_idxs[:, 0], pos_idxs[:, 1]), (num0, num1))
    neg_flat = np.arange(num_total)
    neg_flat = np.delete(neg_flat, pos_flat)
    if not full:
        np.random.shuffle(neg_flat)
        neg_flat = neg_flat[:num_pos]
    neg_idxs = np.array(np.unravel_index(neg_flat, (num0, num1))).T
    neg_atoms0 = non_heavy0.iloc[neg_idxs[:, 0]]
    neg_atoms1 = non_heavy1.iloc[neg_idxs[:, 1]]
    neg_atom_idxs = np.stack((neg_atoms0.index.values, neg_atoms1.index.values)).T
    pos_atom_idxs = np.stack((pos_atoms0.index.values, pos_atoms1.index.values)).T
    return pos_atom_idxs, neg_atom_idxs
